from __future__ import unicode_literals

import functools
import logging
from os.path import join

import mptt
from django.conf import settings
from django.contrib.auth.models import Group
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import (
    DateTimeField,
    ForeignKey as TreeField,
    ManyToManyField,
)
from django.db.models.signals import post_save
from django.urls import reverse_lazy
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _

from touchtechnology.common.db.models import BooleanField
from touchtechnology.common.default_settings import SITEMAP_ROOT
from touchtechnology.common.mixins import NodeRelationMixin

try:
    from django.db.models import JSONField
except ImportError:  # Django 2.2 support
    from django.contrib.postgres.fields import JSONField

logger = logging.getLogger(__name__)


class SitemapNodeBase(models.Model):

    title = models.CharField(max_length=255, verbose_name=_("Title"))

    short_title = models.CharField(
        max_length=100,
        blank=True,
        verbose_name=_("Short title"),
        help_text=_(
            "This is used in navigation menus instead of the longer " "title value."
        ),
    )

    slug = models.SlugField(max_length=255, db_index=True, verbose_name=_("Slug"))

    slug_locked = BooleanField(default=False, verbose_name=_("Slug locked"))

    class Meta:
        abstract = True

    def __str__(self):
        return self.title

    def clean(self):
        if SITEMAP_ROOT is None or self.slug != SITEMAP_ROOT:
            if not self.slug_locked or not self.slug:
                self.slug = slugify(self.title)


@functools.total_ordering
class SitemapNode(NodeRelationMixin, SitemapNodeBase):
    """
    Base class which can be used to represent the structure of the site.
    """

    parent = TreeField(
        "self",
        blank=True,
        null=True,
        verbose_name=_("Parent"),
        related_name="children",
        on_delete=models.PROTECT,
    )

    content_type = models.ForeignKey(
        ContentType, blank=True, null=True, on_delete=models.PROTECT
    )

    object_id = models.PositiveIntegerField(blank=True, null=True)

    object = GenericForeignKey("content_type", "object_id")

    # Migrate from related table in the content application.
    kwargs = JSONField(default=dict)

    require_https = BooleanField(
        default=False,
        verbose_name=_("HTTPS Required"),
        help_text=_("Force this to be served via HTTPS."),
    )

    # visibility and access control
    enabled = BooleanField(
        default=True,
        verbose_name=_("Enabled"),
        help_text=_(
            "Set this to 'No' to disable this object and it's " "children on the site."
        ),
    )

    hidden_from_navigation = BooleanField(
        default=False,
        verbose_name=_("Hide from menus"),
        help_text=_(
            "When set to 'Yes' this object will still be available, "
            "but will not appear in menus."
        ),
    )

    hidden_from_sitemap = BooleanField(
        default=False,
        verbose_name=_("Hide from sitemap"),
        help_text=_(
            "When set to 'Yes' this object will not be listed in "
            "the auto-generated sitemap."
        ),
    )

    hidden_from_robots = BooleanField(
        default=False,
        verbose_name=_("Hide from spiders"),
        help_text=mark_safe(
            _(
                "Set this to <em>Yes</em> to prevent search engines from "
                "indexing this part of the site.<br>"
                "<strong>Warning:</strong> this may affect your ranking "
                "in search engines."
            )
        ),
    )

    restrict_to_groups = ManyToManyField(
        to="auth.Group",
        blank=True,
        verbose_name=_("Restrict to Groups"),
        help_text=_(
            "If you select one or more of these groups your "
            "visitors will need to be logged in and a member of an "
            "appropriate group to view this part of the site."
        ),
    )

    last_modified = DateTimeField(auto_now=True)

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, other):
        if not isinstance(other, SitemapNode):
            return False
        if other is None:
            return False
        return self.pk == other.pk

    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        assert isinstance(other, SitemapNode)
        if other is None:
            return True
        return (self.level, self.lft) < (other.level, other.lft)

    def __repr__(self):
        return '<{0}: "{1}" ({2}:{3},{4})>'.format(
            self.__class__.__name__, self.title, self.level, self.lft, self.rght
        )

    def disable(self):
        logger.debug("Disabling node %r", self)
        self.enabled = False
        return self.save()

    disable.alters_data = True

    def is_home_page(self):
        return self.level == 0 and self.slug == SITEMAP_ROOT

    def is_accessible(self, user):
        groups = Group.objects.filter(sitemapnode=self)
        groups |= Group.objects.filter(sitemapnode__in=self.get_ancestors())
        return set(groups).difference(user.groups.all())

    def get_absolute_url(self):
        parts = [
            ancestor.slug
            for ancestor in self.get_ancestors(include_self=True)
            if not (ancestor.is_root_node() and ancestor.slug == SITEMAP_ROOT)
        ]

        path = "/"

        try:
            path += join(*parts)
        except TypeError:
            logger.debug("SITEMAP_ROOT=%r", self)

        if settings.APPEND_SLASH and not path.endswith("/"):
            path += "/"

        return path

    def _move(self, direction):
        """
        Move a node around within the same level of the tree (ie. parent does
        not change, just the order amongst siblings).
        """
        if direction == "left":
            target = self.get_previous_sibling()
        elif direction == "right":
            target = self.get_next_sibling()
        else:
            raise ValueError(_("direction must be one of 'left' or 'right'"))
        if target is not None:
            self.move_to(target, direction)
            post_save.send(self.__class__, instance=self)

    def move_up(self):
        self._move("left")

    move_up.alters_data = True

    def move_down(self):
        self._move("right")

    move_down.alters_data = True

    @cached_property
    def urls(self):
        # This namespace won't exist unless you are using the
        # touchtechnology.content application in tandem with the
        # touchtechnology.admin application, but you probably won't be using
        # the .urls API if that isn't true anyway!
        try:
            namespace = "admin:content:" + self.content_type.name
        except AttributeError:
            namespace = "admin:content:folder"
        args = (self.id,)
        crud = {
            "edit": reverse_lazy("%s:edit" % namespace, args=args),
            "delete": reverse_lazy("%s:delete" % namespace, args=args),
            "perms": reverse_lazy("%s:perms" % namespace, args=args),
        }
        return crud

    class Meta:
        ordering = ("tree_id", "lft")
        verbose_name = "folder"


mptt.register(SitemapNode)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('competition', '0005_alter_club_facebook_youtube_position'),
    ]

    operations = [
        migrations.AddField(
            model_name='division',
            name='sportingpulse_url',
            field=models.URLField(help_text='Here be dragons! Enter at own risk!', max_length=1024, null=True, blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='match',
            name='external_identifier',
            field=models.CharField(db_index=True, max_length=20, unique=True, null=True, blank=True),
            preserve_default=True,
        ),
    ]



import unittest
from datetime import datetime

from django.test.utils import override_settings
from django.utils import timezone
from freezegun import freeze_time
from test_plus import TestCase
from tournamentcontrol.competition.tests import factories


@override_settings(ROOT_URLCONF="tournamentcontrol.competition.tests.urls")
class GoodViewTests(TestCase):
    def test_index(self):
        self.assertGoodView("competition:index")

    def test_competition(self):
        competition = factories.CompetitionFactory.create()
        self.assertGoodView("competition:competition", competition.slug)

    def test_season(self):
        season = factories.SeasonFactory.create()
        self.assertGoodView("competition:season", season.competition.slug, season.slug)

    def test_season_calendar(self):
        # TODO load matches to see if the query count is scaled properly
        stage = factories.StageFactory.create()
        matches = factories.MatchFactory.create_batch(
            5, stage=stage, date="2022-07-02", time="09:00", datetime="2022-07-02 09:00"
        )
        self.assertGoodView(
            "competition:calendar",
            stage.division.season.competition.slug,
            stage.division.season.slug,
        )

    def test_season_videos(self):
        season = factories.SeasonFactory.create()
        self.assertGoodView(
            "competition:season-videos", season.competition.slug, season.slug
        )

    def test_club(self):
        club = factories.ClubFactory.create()
        stage = factories.StageFactory.create()
        stage.division.season.competition.clubs.add(club)
        team = factories.TeamFactory.create(club=club, division=stage.division)
        factories.MatchFactory.create_batch(stage=stage, home_team=team, size=10)
        self.assertGoodView(
            "competition:club",
            competition=stage.division.season.competition.slug,
            season=stage.division.season.slug,
            club=club.slug,
        )

    def test_club_calendar(self):
        club = factories.ClubFactory.create()
        stage = factories.StageFactory.create(division__season__disable_calendar=True)
        stage.division.season.competition.clubs.add(club)
        team = factories.TeamFactory.create(club=club, division=stage.division)
        factories.MatchFactory.create_batch(stage=stage, home_team=team, size=10)
        self.get(
            "competition:calendar",
            competition=stage.division.season.competition.slug,
            season=stage.division.season.slug,
            club=club.slug,
        )
        self.response_410()

    def test_division(self):
        division = factories.DivisionFactory.create()
        self.assertGoodView(
            "competition:division",
            division.season.competition.slug,
            division.season.slug,
            division.slug,
        )

    def test_division_calendar(self):
        division = factories.DivisionFactory.create()
        self.assertGoodView(
            "competition:calendar",
            competition=division.season.competition.slug,
            season=division.season.slug,
            division=division.slug,
        )

    def test_stage(self):
        stage = factories.StageFactory.create()
        self.assertGoodView(
            "competition:stage",
            stage.division.season.competition.slug,
            stage.division.season.slug,
            stage.division.slug,
            stage.slug,
        )

    def test_stage_group(self):
        pool = factories.StageGroupFactory.create()
        self.assertGoodView(
            "competition:pool",
            pool.stage.division.season.competition.slug,
            pool.stage.division.season.slug,
            pool.stage.division.slug,
            pool.stage.slug,
            pool.slug,
        )

    def test_match(self):
        match = factories.MatchFactory.create()
        self.assertGoodView(
            "competition:match",
            match.stage.division.season.competition.slug,
            match.stage.division.season.slug,
            match.stage.division.slug,
            match.pk,
        )

    def test_match_gone(self):
        opts = [
            {"is_bye": True},
            {"home_team": None},
            {"away_team": None},
        ]
        for kw in opts:
            match = factories.MatchFactory.create(**kw)
            self.get(
                "competition:match",
                match.stage.division.season.competition.slug,
                match.stage.division.season.slug,
                match.stage.division.slug,
                match.pk,
            )
            self.response_410()

    def test_match_video(self):
        match = factories.MatchFactory.create(videos=["https://youtu.be/jNQXAC9IVRw"])
        self.assertGoodView(
            "competition:match-video",
            match.stage.division.season.competition.slug,
            match.stage.division.season.slug,
            match.stage.division.slug,
            match.pk,
        )
        self.assertResponseContains(
            '<iframe width="480" height="360" '
            'src="http://www.youtube.com/embed/jNQXAC9IVRw?wmode=opaque" '
            'loading="lazy" frameborder="0" allowfullscreen>'
            "</iframe>"
        )

    def test_match_video_gone(self):
        opts = [
            {"is_bye": True},
            {"home_team": None},
            {"away_team": None},
        ]
        for kw in opts:
            match = factories.MatchFactory.create(**kw)
            self.get(
                "competition:match-video",
                match.stage.division.season.competition.slug,
                match.stage.division.season.slug,
                match.stage.division.slug,
                match.pk,
            )
            self.response_410()


@override_settings(ROOT_URLCONF="tournamentcontrol.competition.tests.urls")
class FrontEndTests(TestCase):
    def test_competition_list(self):
        comp_1 = factories.CompetitionFactory.create()
        comp_2 = factories.CompetitionFactory.create(enabled=False)
        self.get("competition:index")
        self.assertResponseContains(comp_1.title, html=False)
        self.assertResponseNotContains(comp_2.title, html=False)

    def test_season_list(self):
        season = factories.SeasonFactory.create()
        self.assertGoodView("competition:season", season.competition.slug, season.slug)
        self.assertResponseContains(season.title, html=False)

        season = factories.SeasonFactory.create(competition__enabled=False)
        self.get("competition:competition", season.competition.slug)
        self.response_404()

    def test_division_list(self):
        division = factories.DivisionFactory.create()
        self.assertGoodView(
            "competition:season", division.season.competition.slug, division.season.slug
        )
        self.assertResponseContains(division.title, html=False)

    def test_division_view(self):
        division = factories.DivisionFactory.create()
        self.assertGoodView(
            "competition:division",
            division.season.competition.slug,
            division.season.slug,
            division.slug,
        )
        self.assertResponseContains(division.title, html=False)

    @freeze_time("2013-11-01")
    def test_upcoming_matches(self):
        division = factories.DivisionFactory.create()
        factories.MatchFactory.create_batch(
            stage__division=division,
            datetime=datetime(2013, 11, 22, 10, tzinfo=timezone.utc),
            size=10,
        )
        self.assertGoodView(
            "competition:season", division.season.competition.slug, division.season.slug
        )
        self.assertResponseContains("Nov. 22, 2013", html=False)

    def test_division_match_list(self):
        division = factories.DivisionFactory.create()
        factories.MatchFactory.create_batch(stage__division=division, size=10)
        self.assertGoodView(
            "competition:division",
            division.season.competition.slug,
            division.season.slug,
            division.slug,
        )
        for team in division.teams.all():
            href = self.reverse(
                "competition:team",
                division.season.competition.slug,
                division.season.slug,
                division.slug,
                team.slug,
            )
            self.assertResponseContains(
                '<a href="{0}">{1}</a>'.format(href, team.title)
            )

    def test_team_calendar(self):
        team = factories.TeamFactory.create()
        factories.MatchFactory.create_batch(
            stage__division=team.division, home_team=team, size=5
        )
        self.assertGoodView(
            "competition:calendar",
            team.division.season.competition.slug,
            team.division.season.slug,
            team.division.slug,
            team.slug,
        )
        for opponent in team.division.teams.exclude(pk=team.pk):
            subject = "{} vs {}".format(team.title, opponent.title)
            self.assertResponseContains(subject, html=False)

    @unittest.expectedFailure
    def test_team_calendar_disabled(self):
        season = factories.SeasonFactory.create(disable_calendar=True)
        team = factories.TeamFactory.create(division__season=season)
        factories.MatchFactory.create_batch(
            stage__division=team.division, home_team=team, size=5
        )
        self.assertGoodView(
            "competition:calendar",
            team.division.season.competition.slug,
            team.division.season.slug,
            team.division.slug,
            team.slug,
        )

    def test_division_calendar(self):
        team = factories.TeamFactory.create()
        factories.MatchFactory.create_batch(
            stage__division=team.division, home_team=team, size=5
        )
        self.assertGoodView(
            "competition:calendar",
            team.division.season.competition.slug,
            team.division.season.slug,
            team.division.slug,
        )
        for opponent in team.division.teams.exclude(pk=team.pk):
            subject = "{} vs {}".format(team.title, opponent.title)
            self.assertResponseContains(subject, html=False)

    @unittest.expectedFailure
    def test_division_calendar_disabled(self):
        season = factories.SeasonFactory.create(disable_calendar=True)
        team = factories.TeamFactory.create(division__season=season)
        factories.MatchFactory.create_batch(
            stage__division=team.division, home_team=team, size=5
        )
        self.assertGoodView(
            "competition:calendar",
            team.division.season.competition.slug,
            team.division.season.slug,
            team.division.slug,
        )

    def test_season_calendar(self):
        team = factories.TeamFactory.create()
        factories.MatchFactory.create_batch(
            stage__division=team.division, home_team=team, size=5
        )
        self.assertGoodView(
            "competition:calendar",
            team.division.season.competition.slug,
            team.division.season.slug,
        )
        for opponent in team.division.teams.exclude(pk=team.pk):
            subject = "{} vs {}".format(team.title, opponent.title)
            self.assertResponseContains(subject, html=False)

    @unittest.expectedFailure
    def test_season_calendar_disabled(self):
        season = factories.SeasonFactory.create(disable_calendar=True)
        team = factories.TeamFactory.create(division__season=season)
        factories.MatchFactory.create_batch(
            stage__division=team.division, home_team=team, size=5
        )
        self.assertGoodView(
            "competition:calendar",
            team.division.season.competition.slug,
            team.division.season.slug,
        )



import logging
import operator
import os
import re
import socket
from decimal import Decimal
from itertools import islice, zip_longest
from urllib.parse import parse_qsl

import pkg_resources
from classytags.arguments import Argument
from classytags.core import Options
from classytags.helpers import AsTag
from django.conf import settings
from django.db.models import Model, Q
from django.db.models.query import QuerySet
from django.forms.boundfield import BoundField
from django.forms.widgets import (
    CheckboxInput,
    CheckboxSelectMultiple,
    Input,
    MultiWidget,
    RadioSelect,
    Select,
    Textarea,
)
from django.template.base import Node
from django.template.library import Library
from django.template.loader import get_template, render_to_string
from django.urls import Resolver404, resolve, reverse
from django.utils import timezone
from django.utils.encoding import smart_str
from django.utils.html import escape
from django.utils.http import urlencode
from django.utils.safestring import mark_safe
from django.utils.text import slugify
from guardian.core import ObjectPermissionChecker
from namedentities import named_entities

from touchtechnology.common.default_settings import CURRENCY_SYMBOL
from touchtechnology.common.exceptions import NotModelManager
from touchtechnology.common.models import SitemapNode
from touchtechnology.common.utils import (
    create_exclude_filter,
    get_all_perms_for_model_cached,
    model_and_manager,
    tree_for_node,
)
from tournamentcontrol.competition.utils import FauxQueryset

logger = logging.getLogger(__name__)

camel_case_re = re.compile(r"(?P<new_word>((?<![A-Z])[A-Z]|[A-Z](?![A-Z0-9])))")

version_re = re.compile(
    r"""^
    (?P<version>\d+\.\d+)         # minimum 'N.N'
    (?P<extraversion>(?:\.\d+)*)  # any number of extra '.N' segments
    (?:
        (?P<prerel>[abc]|rc)         # 'a' = alpha, 'b' = beta
                                     # 'c' or 'rc' = release candidate
        (?P<prerelversion>\d+(?:\.\d+)*)
    )?
    (?P<postdev>(\.post(?P<post>\d+))?(\.dev(?P<dev>\d+))?)?
    $""",
    re.VERBOSE,
)

FORM_FIELD_TEMPLATE = get_template("touchtechnology/common/templatetags/field.html")

register = Library()


@register.filter
def at_a_time(i, n):
    """
    Turn a list into smaller lists of n items. For example, if we have:

        object_list = [1, 2, 3, 4, 5, 6, 7, 8]

    The {% object_list|at_a_time:2 %} will result in:

        [[1, 2], [3, 4], [5, 6], [7, 8]

    Or {% object_list|at_a_time:3 %} will result in:

        [[1, 2, 3], [4, 5, 6], [7, 8, None]]

    We don't filter out None items, so check when iterating in your template.
    """
    return zip_longest(*[islice(i, x, None, n) for x in range(n)])


@register.filter
def camel_case_to_underscores(s):
    """Converts a CamelCase name to use lowercase_and_underscores.

    e.g. "ConvertHTMLToText" -> "convert_html_to_text"
    """

    def convert(match):
        content = match.group().lower()
        if match.start() == 0:
            return content
        else:
            return "_%s" % content

    return camel_case_re.sub(convert, s).lower()


@register.filter
def camel_case_split(s):
    """Converts a CamelCase name to be split by spaces.

    e.g. "CovertHTMLToText" -> "Convert HTML To Text"
    """

    def convert(match):
        content = match.group()
        if match.start() == 0:
            return content
        else:
            return " %s" % content

    return camel_case_re.sub(convert, s)


@register.filter
def cssify(s):
    return slugify(s or "").replace("-", "_")


@register.filter
def future(d):
    if d > timezone.now():
        return True
    return False


@register.filter
def split(st, on):
    return st.split(on)


@register.filter
def twittify(s):
    twitter_re = re.compile(r"(@(?P<username>[a-z0-9_]+)?)", re.I)
    return mark_safe(
        twitter_re.sub(
            r'@<a class="twitter user" '
            'target="_blank" '
            'href="http://twitter.com/\\2">\\2</a>',
            s,
        )
    )


@register.filter
def count(queryset):
    return queryset.count()


@register.filter
def disabled(queryset):
    return queryset.exclude(enabled=True)


@register.filter
def enabled(queryset):
    return queryset.filter(enabled=True)


@register.filter
def invisible(queryset):
    return queryset.exclude(hidden_from_navigation=False)


@register.filter
def visible(queryset):
    return queryset.filter(hidden_from_navigation=False)


@register.tag
def navigation(parser, token):
    """
    Build a navigational structure.

        Examples:

        {% navigation %}
        {% navigation current_node=node %}
        {% navigation start_at=1 stop_at=3 %}
        {% navigation root='page_3' %}
    """
    args = token.split_contents()[1:]
    kwargs = {k: parser.compile_filter(v) for k, v in [a.split("=", 1) for a in args]}
    return NavigationNode(kwargs)


class NavigationNode(Node):
    def __init__(self, kwargs):
        self.kwargs = kwargs

    def render(self, context):
        r_kwargs = dict(
            [
                (str(k), v.resolve(context, ignore_failures=True))
                for k, v in self.kwargs.items()
            ]
        )
        return do_navigation(**r_kwargs)


def do_navigation(
    root=None,
    start_at=None,
    stop_at=None,
    current_node=None,
    expand_all_nodes=None,
    template_name=None,
    **kwargs,
):
    nodes = SitemapNode._tree_manager.select_related("content_type", "parent")

    if template_name is None:
        template_name = "touchtechnology/common/templatetags/navigation.html"

    if expand_all_nodes is None:
        expand_all_nodes = False

    logger.debug("======== do_navigation ========")
    logger.debug("template_name: %r", template_name)
    logger.debug("expand_all_nodes: %r", expand_all_nodes)
    logger.debug("current_node: %r", current_node)

    if root is not None:
        if isinstance(root, str):
            try:
                root = resolve(root).kwargs.get("node")
            except Resolver404:
                root = resolve(reverse(root)).kwargs.get("node")

        logger.debug("root: %r", root)

        try:
            nodes = root.get_descendants(include_self=True)
        except AttributeError:
            nodes = SitemapNode.objects.none()

        if (
            current_node is not None
            and current_node.tree_id == root.tree_id
            and root.lft < current_node.lft
            and not expand_all_nodes
        ):
            nodes = tree_for_node(current_node)

    elif current_node is not None and not expand_all_nodes:
        nodes = tree_for_node(current_node)

    logger.debug("nodes: %r", nodes)

    # make sure we hide any nodes that are in a hidden part of the tree
    nodes_hidden_from_navigation = nodes.filter(
        Q(hidden_from_navigation=True) | Q(enabled=False)
    )

    hidden_from_navigation = create_exclude_filter(nodes_hidden_from_navigation)

    nodes = nodes.exclude(hidden_from_navigation)

    logger.debug("nodes[cleaned]: %r", nodes)

    # flatten the list of nodes to a list
    tree = list(nodes)

    if current_node is None and not expand_all_nodes:
        stop_at = max(start_at or 0, stop_at or 0, 0)

    if start_at is not None:
        tree = [n for n in tree if n.level >= start_at]

    if stop_at is not None:
        tree = [n for n in tree if n.level <= stop_at]

    if not expand_all_nodes and current_node is not None:
        parents = []
        n = current_node
        while n.parent is not None:
            parents.append(n.parent)
            n = n.parent

        fmt = "[{rel}] {url} {node}"

        def func(node):
            """
            Filter function to determine if a node should appear in the
            navigation tree or not.
            """
            rel = current_node.rel(node)
            url = node.get_absolute_url()
            logger.debug(fmt.format(node=node, rel=rel, url=url))
            return rel in {
                "ROOT",
                "ANCESTOR",
                "PARENT",
                "UNCLE",
                "ME",
                "SIBLING",
                "DESCENDANT",
            }

        log = {
            "rel": "NODE",
            "node": repr(current_node),
            "url": current_node.get_absolute_url(),
        }
        logger.debug(fmt.format(**log))
        tree = [t for t in tree if func(t)]

    # re-sort the queryset to get our correct tree structure back
    tree = sorted(tree, key=operator.attrgetter("tree_id", "lft"))

    context = {
        "nodes": tree,
        "current_node": current_node,
        "hidden_nodes": nodes_hidden_from_navigation,
        "start_at": start_at,
        "stop_at": stop_at,
    }

    return render_to_string(template_name, context)


@register.simple_tag
def field(bf, label=None):
    if not isinstance(bf, BoundField):
        raise TypeError(
            "{{% field %}} tag can only be used with " "BoundFields ({0})".format(bf)
        )

    if bf.is_hidden:
        return smart_str(bf)

    widget = bf.field.widget
    widget_class_name = camel_case_to_underscores(widget.__class__.__name__)

    if label is None:
        label = bf.label

    if isinstance(widget, (CheckboxInput,)):
        radio_checkbox_input = True
    else:
        radio_checkbox_input = False

    if label:
        if isinstance(widget, (Input, Select, Textarea),) and not isinstance(
            widget, (CheckboxInput, RadioSelect, CheckboxSelectMultiple, MultiWidget)
        ):
            # Use a <label> tag
            caption = bf.label_tag(label, attrs={"class": "field_name"})
        else:
            # Don't use a <label> tag
            label_suffix = bf.form.label_suffix or ""
            caption = f'<span class="field_name">{label}{label_suffix}</span>'
    else:
        caption = ""

    context = {
        "f": bf,
        "caption": caption,
        "widget_class_name": widget_class_name,
        "radio_checkbox_input": radio_checkbox_input,
        "no_label": not bool(label),
    }

    return FORM_FIELD_TEMPLATE.render(context)


@register.inclusion_tag("touchtechnology/common/templatetags/analytics.html")
def analytics(code=None):
    if code is None:
        code = getattr(settings, "GOOGLE_ANALYTICS", code)
    context = {
        "code": code,
        "debug": not os.environ.get("SITE_ENV", "dev") == "live",
    }
    return context


@register.inclusion_tag(
    "touchtechnology/common/templatetags/pagination.html", takes_context=True
)
def pagination(context):
    query_string = context.get("QUERY_STRING", "")
    query = parse_qsl(query_string, True)
    query = [q for q in query if q[0] != "page"]
    query_string = urlencode(query, True)
    context.update({"QUERY_STRING": query_string})
    return context


@register.filter("type")
def get_type(obj):
    if isinstance(obj, QuerySet):
        return obj.model._meta.verbose_name
    elif isinstance(obj, Model):
        return obj._meta.verbose_name
    else:
        return obj.__class__.__name__


@register.filter("types")
def get_type_plural(obj):
    if isinstance(obj, (QuerySet, FauxQueryset)):
        return obj.model._meta.verbose_name_plural
    elif isinstance(obj, Model):
        return obj._meta.verbose_name_plural
    else:
        # extremely naive
        return obj.__class__.__name__ + "s"


@register.filter
def htmlentities(s):
    replaced_entities = named_entities(
        escape(s).encode("ascii", "xmlcharrefreplace").decode("utf8")
    )
    return mark_safe(replaced_entities)


@register.filter("abs")
def absolute_value(value):
    return abs(value)


@register.inclusion_tag("touchtechnology/common/templatetags/price.html")
def price(value, extra=""):
    context = {
        "SYMBOL": CURRENCY_SYMBOL,
        "AMOUNT": value or Decimal("0.00"),
        "EXTRA": extra,
    }
    return context


@register.filter("islice", is_safe=True)
def islice_(value, arg):
    """
    Returns an iterator slice of the list.
    """
    try:
        bits = []
        for x in arg.split(":"):
            if len(x) == 0:
                bits.append(None)
            else:
                bits.append(int(x))
        return islice(value, *bits)

    except (ValueError, TypeError):
        return value  # Fail silently.


@register.inclusion_tag("touchtechnology/common/templatetags/version.html")
def version(package, url=None):
    environment = pkg_resources.Environment()
    context = {}
    try:
        if package == "python":
            context = {
                "name": "Python",
                "project": "Python",
                "version": environment.python,
                "release": None,
            }
        else:
            distribution = environment[package][-1]
            release = version_re.match(distribution.version).groupdict()
            context = {
                "name": distribution.key,
                "project": distribution.project_name,
                "version": distribution.version,
                "release": release.get("prerel"),
            }
    except IndexError:
        logger.exception(
            'Attempting to lookup version for package "%s" not ' "in pkg_resources",
            package,
        )
    if url:
        context["url"] = url
    return context


@register.filter("permchecker")
def create_permission_checker(model_or_manager, user):
    """
    Create an :class:`ObjectPermissionChecker` for optimal permission checking
    of related objects.

    http://django-guardian.readthedocs.io/en/stable/userguide/performance.html
    """
    model, manager = model_and_manager(model_or_manager)
    checker = ObjectPermissionChecker(user)
    try:
        checker.prefetch_perms(manager)
    except UnboundLocalError:
        logger.exception(
            "https://github.com/django-guardian/django-guardian/issues/519"
        )
    return checker


@register.filter("checkperm")
def check_permission(obj, checker):
    """
    For a given model instance and a :class:`ObjectPermissionChecker` return
    which permissions the associated user has.
    """
    return [
        model_perm
        for model_perm in get_all_perms_for_model_cached(obj._meta.model)
        if checker.has_perm(model_perm, obj)
    ]


@register.filter("hasperm")
def has_permission(obj, user):
    try:
        model, manager = model_and_manager(obj)
    except NotModelManager:
        logger.exception('error="model cannot be determined"')
        return set()

    logger.debug(
        'model="%s.%s", object_id="%s", user="%s"',
        model._meta.app_label,
        model._meta.model_name,
        getattr(obj, "pk", ""),
        user.get_username(),
    )

    # Calculate the permissions this user has for the given object, both
    # directly and inferred by group memberships.
    perms = user.get_all_permissions(obj)
    perms |= user.get_group_permissions(obj)

    # What are the permissions that this model accepts?
    model_perms = get_all_perms_for_model_cached(model, ttl=300)

    # Superusers have all permissions, so add each permission that the model
    # accepts for this user to the set that has been explicitly cast.
    if user.is_superuser:
        perms.update(model_perms)

    # Otherwise we need to iterate the groups that the user belongs to and see
    # if they transfer permission to the user.
    else:
        for group in user.groups.all():
            group_perms = group.permissions.all()
            group_model_perms = [p.codename for p in model_perms if p in group_perms]
            logger.debug(
                'user="%s", group="%s", permissions="%s"',
                user.get_username(),
                group,
                ", ".join(group_model_perms),
            )
            perms.update(group_model_perms)

    # Log the resulting set of permissions for this user for this object.
    logger.debug('user="%s", permissions=%r', user.get_username(), perms)

    return perms


@register.simple_tag
def host():
    return socket.gethostname()


@register.inclusion_tag("touchtechnology/common/templatetags/hostname.html")
def hostname():
    hostname = socket.gethostname()
    host, domain = hostname.split(".", 1)
    return dict(hostname=hostname, host=host, domain=domain)


class Login(AsTag):
    options = Options("as", Argument("varname", required=False, resolve=False))

    def get_value(self, context):
        request = context.get("request")
        url = reverse("accounts:login")
        if request:
            url += "?next=" + request.path
        return url


register.tag(Login)



import collections
import logging
import re
from datetime import datetime
from decimal import Decimal
from operator import and_, attrgetter, or_

import pytz
from django.apps import apps
from django.contrib.auth import get_user_model
from django.db.models import F, Q
from django.http import HttpResponse
from django.template.loader import select_template
from django.utils import timezone
from touchtechnology.common.prince import prince

try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest


logger = logging.getLogger(__name__)

home_team_needs_progressing = and_(
    Q(home_team__isnull=True),
    or_(Q(home_team_undecided__isnull=False), Q(home_team_eval__isnull=False)),
)

away_team_needs_progressing = and_(
    Q(away_team__isnull=True),
    or_(Q(away_team_undecided__isnull=False), Q(away_team_eval__isnull=False)),
)

team_needs_progressing = and_(
    or_(home_team_needs_progressing, away_team_needs_progressing),
    Q(stage__division__season__complete=False),
)

legitimate_bye_match = and_(
    Q(is_bye=True), or_(Q(home_team__isnull=False), Q(away_team__isnull=False))
)

match_unplayed = or_(Q(home_team_score__isnull=True), Q(away_team_score__isnull=True))

stage_group_position_re = re.compile(
    r"""
    (?:
        S(?P<stage>\d+)         # stage number (optional)
    )?
    (?:
        G(?P<group>\d+)         # group number (optional)
    )?
    P(?P<position>\d+)          # position number
    """,
    re.VERBOSE,
)


class FauxQueryset(list):
    """
    In some scenarios we need to construct a list-like object so we
    can handle the ordering of the objects.
    """

    def __init__(self, model, team=None, *args, **kwargs):
        super(FauxQueryset, self).__init__(*args, **kwargs)
        # the team attribute is a legacy from a previous implementation
        # detail, we need to improve this. FIXME.
        self.team = team
        self.model = model
        self.ordered = True
        self.db = None
        self._prefetch_related_lookups = True  # ModelChoiceIterator

    def all(self):
        return self

    def get(self, **kwargs):
        pks = [each.pk for each in self]
        return self.model.objects.filter(pk__in=pks).get(**kwargs)


class SumDict(dict):
    """
    When calculating the values for a LadderSummary we may need to
    quickly add together aggregated values. These will already be
    accessible as a `dict` so this will quickly allow us to add the
    values together and get a total.
    """

    def __add__(self, other):
        i = {}
        for d in (self, other):
            for k, v in d.items():
                i.setdefault(k, []).append(v if v is not None else 0)
        return SumDict([(k, sum(v)) for k, v in i.items()])


#
# Mathematical
#


def ceiling(value, factor=1):
    """
    For a Decimal value, round it up to the nearest multiple of ``factor``.

        >>> ceiling(3.5)
        Decimal('4')
        >>> ceiling(3.5, 0.75)
        Decimal('3.75')

    :param value: number to be rounded up
    :param factor: multiplier to round up to
    :return: Decimal
    """
    value = Decimal(str(value))
    factor = Decimal(str(factor))
    return (value / factor).quantize(1, rounding="ROUND_UP") * factor


def floor(value, factor=1):
    """
    For a Decimal value, round it down to the nearest multiple of ``factor``.

        >>> floor(3.5)
        Decimal('3')
        >>> floor(2.5, 0.6)
        Decimal('2.4')

    :param value: number to be rounded down
    :param factor: multiplier to round down to
    :return: Decimal
    """
    value = Decimal(str(value))
    factor = Decimal(str(factor))
    return (value / factor).quantize(1, rounding="ROUND_DOWN") * factor


def revpow(n, base):
    """
    Reverse of ``pow`` built-in function.

        >>> for i in range(4):
        ...     revpow(pow(2, i), 2) == i
        ...
        True
        True
        True
        True

    :param n: number that is a power of base
    :param base: the base
    :return: int
    """
    res = 0
    intermediate = n // base
    if intermediate:
        res += 1
        res += revpow(intermediate, base)
    return res


#
# Iterables
#


def grouper(iterable, size, fill=None):
    """
    Collect data into fixed-length chunks or blocks. Lifted from
    https://docs.python.org/2/library/itertools.html#recipes

        >>> list(grouper('ABCDEFG', 3, 'x'))
        [('A', 'B', 'C'), ('D', 'E', 'F'), ('G', 'x', 'x')]

    :param iterable: thing to be broken into chunks
    :param size: size of each chunk
    :param fill: value to use to pad chunks of less than size
    :return: zip_longest
    """
    args = [iter(iterable)] * size
    return zip_longest(fillvalue=fill, *args)


#
# Timeslots & timezones
#


def combine_and_localize(date, time, tz):
    """
    Given the date and time parts, return a timezone aware
    datetime instance. Date and time parts are considered
    local to tzinfo.
    """
    combined = datetime.combine(date, time)
    if isinstance(tz, str):
        tz = pytz.timezone(tz)
    return timezone.make_aware(combined, tz)


def time_choice(t):
    return (t.strftime("%H:%M:%S"), t.strftime("%H:%M"))


#
# Competition draw generation utilities
#


def round_robin(teams, rounds=None):
    """
    Based on recipe found at http://code.activestate.com/recipes/65200/
    """
    # ensure we have a list so we can perform appends and inserts
    if not isinstance(teams, list):
        teams = [t for t in teams]

    # ensure we have an even number of teams
    if len(teams) % 2:
        teams.append(0)

    count = len(teams)
    half = int(count / 2)

    # if the rounds is not set, we will produce one complete round
    if rounds is None:
        rounds = count - 1

    schedule = []
    for turn in range(rounds):
        pairings = []
        for i in range(half):
            pairings.append((teams[i], teams[count - i - 1]))
        teams.insert(1, teams.pop())
        schedule.append(pairings)

    return schedule


def round_robin_format(teams, rounds=None):
    i, s = 1, ""
    for round in round_robin(teams, rounds):
        s += "ROUND\n"
        for home, away in round:
            s += "%s: %s vs %s\n" % (i, home, away)
            i += 1
    return s.strip()


def final_series_rounds(pools):
    """
    For a given number of pools determine the correct number of rounds to
    play in a final series to fairly determine a winner.

    :param pools: number of pools
    :return: int
    """
    rounds = revpow(pools, 2)
    if rounds <= 1:
        return 2
    return rounds + 1


def final_series_round_label(number_of_matches):
    if number_of_matches >= 8:
        return "Round of %d" % (number_of_matches * 2)
    if number_of_matches == 4:
        return "Quarter Final"
    if number_of_matches == 2:
        return "Semi Final"
    return "Final"


def single_elimination_final_format(number_of_pools, bronze_playoff=None):
    """
    For any number of pools, construct a single elimination draw format to
    find the champion team.

    :param number_of_pools: number of starting groups
    :param bronze_playoff: unicode label for 3rd place playoff. If None no
                           match will be scheduled.
    :return: list
    """
    from tournamentcontrol.competition.draw import (
        MatchDescriptor,
        RoundDescriptor,
    )

    # Start building our final series with the initial round. Each match is
    # ordered so that the pool from which the highest seeds were originally
    # placed (pool 1) are farthest away from the second highest seed (pool 2).
    # In effect, this means P1, P2, ... PN.

    if number_of_pools == 1:
        initial = RoundDescriptor(2, final_series_round_label(2))
        initial.add(MatchDescriptor(1, "P1", "P4"))
        initial.add(MatchDescriptor(2, "P2", "P3"))

    else:
        initial = RoundDescriptor(
            number_of_pools, final_series_round_label(number_of_pools)
        )

        for pool in range(number_of_pools):
            match = MatchDescriptor(
                pool + 1, "G%dP1" % (pool + 1), "G%dP2" % (number_of_pools - pool),
            )
            initial.add(match)

    series = [
        initial,
    ]

    # Until we reach the final round (we don't know how many hops yet),
    # introspect the previous round and add a match between plays the highest
    # remaining seed against the lowest remaining seed.
    # In effect, this means W1 vs W-1, W2 vs W-2, ...

    while len(series[-1].matches) > 1:
        # Always half as many as the previous round, we're eliminating teams!
        matches_this_round = len(series[-1].matches) // 2

        this_round = RoundDescriptor(
            matches_this_round, final_series_round_label(matches_this_round)
        )

        for i in range(matches_this_round):
            match = MatchDescriptor(
                series[-1].matches[-1].match_id + i + 1,
                "W%s" % (series[-1].matches[i].match_id),
                "W%s" % (series[-1].matches[-1 - i].match_id),
            )
            this_round.add(match)

        series.append(this_round)

    # If we want to schedule a 3rd place playoff, add an extra match to the
    # final round we've already created.
    if bronze_playoff is not None:
        match = MatchDescriptor(
            series[-1].matches[0].match_id + 1,
            "L%s" % (series[-2].matches[0].match_id),
            "L%s" % (series[-2].matches[1].match_id),
            bronze_playoff,
        )
        series[-1].add(match)

        # Also update the match label of the only other match in this stage.
        series[-1].matches[0].match_label = series[-1].round_label

    return series


def stage_group_position(stage, formula):
    match = stage_group_position_re.match(formula)
    if not match:
        logger.exception("Formula %r is invalid", formula)
        raise ValueError("Invalid formula")

    s, g, p = match.groups()
    logger.debug("stage = %s, group = %s, position = %s", s, g, p)

    if s is None:
        logger.debug("No stage specified in formula %r", formula)
        s = stage.comes_after
    else:
        try:
            s = stage.division.stages.all()[int(s) - 1]
        except IndexError:
            logger.exception("Invalid stage %r for division %r", s, stage.division)
            raise
    logger.debug("stage=%r", s)

    if g is None:
        logger.debug("No group specified in formula %r", formula)
    else:
        try:
            g = s.pools.all()[int(g) - 1]
        except IndexError:
            logger.exception("Invalid group %r for stage %r", g, stage)
            raise
    logger.debug("group=%r", g)

    p = int(p)
    logger.debug("position=%r", p)

    return s, g, p


#
# Scorecards
#


def generate_scorecards(
    matches=None,
    templates=None,
    format="html",
    extra_context=None,
    stage=None,
    **kwargs
):
    if extra_context is None:
        extra_context = {}

    context = {
        "matches": matches,
    }
    context.update(extra_context)

    template = select_template(templates)
    output = template.render(context)

    if format == "pdf":
        output = prince(output, **kwargs)

    if stage is not None:
        for match in matches:
            match.to_be_printed.remove(stage)

    return output


def generate_fixture_grid(
    season,
    dates=None,
    templates=None,
    format="html",
    request=None,
    extra_context=None,
    **kwargs
):
    if dates is None:
        dates = season.matches.dates("date", "day")

    if templates is None:
        templates = ["tournamentcontrol/competition/admin/grid.html"]

    if extra_context is None:
        extra_context = {}

    matrices = collections.OrderedDict()

    play_at = season.get_places()

    for date in dates:
        matches = season.matches.select_related(
            "stage_group",
            "stage",
            "stage__division",
            "home_team",
            "home_team__club",
            "away_team",
            "away_team__club",
        ).filter(date=date)

        times = sorted(
            {m.time for m in matches if m.time is not None}.union(
                season.get_timeslots(date)
            )
        )

        keyed = collections.defaultdict(lambda: None)
        for m in matches:
            keyed.setdefault((m.play_at, m.time), []).append(m)

        matrix = collections.OrderedDict()
        for t in times:
            matrix.setdefault(t, collections.OrderedDict())
            for p in play_at:
                matrix[t].setdefault(p, keyed[(p, t)])

        matrices.setdefault(date, matrix)

    context = {
        "matrices": matrices,
        "venues": season.venues.all(),
        "columns": play_at,
    }
    context.update(extra_context)

    template = select_template(templates)
    html = template.render(context)

    if format == "pdf":
        pdf = prince(html, **kwargs)
        response = HttpResponse(pdf, content_type="application/pdf")
    else:
        response = HttpResponse(html)

    return response


#
# Forfeit handling
#


def forfeit_notification_recipients(match):
    """
    Return a 2-tuple of users to contact advising them that a forfeit
    notification has been received from team for match.

    The first element is a queryset of the affected players, while the second
    is the queryset of users attached to the matches season to be notified.

    This function does not send any notifications!
    """
    UserModel = get_user_model()

    people = match.forfeit_winner.people.exclude(person__user__isnull=True)
    user = UserModel.objects.filter(person__in=people.values("person"))

    return (user, match.stage.division.season.forfeit_notifications.all())


#
# regrade
#


def regrade(team, to, from_date=None):
    Division = apps.get_model("competition", "Division")
    Team = apps.get_model("competition", "Team")

    assert isinstance(team, Team), "team is not a Team instance"
    assert isinstance(to, Division), "to is not a Division instance"
    assert team.division != to, "move to same Division is not allowed"
    assert team.division.season == to.season, "move to different Season is not allowed"

    # Work out the regrade date if not specified
    if from_date is None:
        from_date = team.matches.filter(match_unplayed).earliest("date").date

    # Update the sequence numbers of all teams in the current division to keep
    # them in order without any gaps
    team.division.teams.filter(order__gt=team.order).update(order=F("order") - 1)

    # All unplayed matches that this team is assigned to play need to be turned
    # into byes, removing them from the home_team and away_team fields. Strip
    # the time and field also.
    old_matches = team.matches.filter(match_unplayed, date__gte=from_date)
    old_matches.filter(home_team=team).update(
        home_team=None, is_bye=True, time=None, datetime=None, play_at=None
    )
    old_matches.filter(away_team=team).update(
        away_team=None, is_bye=True, time=None, datetime=None, play_at=None
    )

    # Move team into the bye matches in the new division.
    new_matches = to.matches.filter(legitimate_bye_match, date__gte=from_date)
    new_matches.filter(home_team=None).update(home_team=team, is_bye=False)
    new_matches.filter(away_team=None).update(away_team=team, is_bye=False)

    # Determine the highest sequence value in the target division and assign
    # that to our team. If it throws an DoesNotExist exception, the division
    # is empty and we need to set it to 1.
    try:
        team.order = to.teams.latest("order").order + 1
    except Team.DoesNotExist:
        team.order = 1


#
# label_from_instance utilities
#


def team_and_division(o):
    "For use on Model.team_clashes as label_from_instance argument."
    return "%s (%s)" % (o.title, o.division.title)



from __future__ import unicode_literals

from operator import itemgetter

import pytz
from django import forms
from django.utils import timezone
from django.utils.translation import gettext_lazy as _


def timezone_choice(tzname, country=None):
    """
    Given a timezone string, convert it to a two-tuple suitable
    for a choice in a FormField.

    >>> timezone_choice('UTC')
    ('UTC', 'UTC')
    >>> timezone_choice('Australia/Sydney')
    ('Australia/Sydney', 'Sydney')
    >>> timezone_choice('America/Indiana/Indianapolis')
    ('America/Indiana/Indianapolis', 'Indianapolis (Indiana)')
    """
    try:
        base, rest = tzname.split("/", 1)
    except ValueError:
        base, rest = None, tzname
    rest = rest.replace("_", " ")
    try:
        mid, rest = rest.split("/")
        if mid != base and mid != country:
            rest = "{0} ({1})".format(rest, mid)
    except ValueError:
        pass
    return (tzname, rest)


def timezone_choices(countries, timezones, key):
    yield "", ""
    for iso3166, country in sorted(countries.items(), key=key):
        zones = timezones.get(iso3166, [])
        zones = [timezone_choice(zone, country) for zone in zones]
        zones.sort(key=key)
        yield country, zones


DAY_CHOICES = [("", "")] + list(zip(*[range(1, 32)] * 2))

MONTH_CHOICES = (
    ("", ""),
    (1, "January"),
    (2, "February"),
    (3, "March"),
    (4, "April"),
    (5, "May"),
    (6, "June"),
    (7, "July"),
    (8, "August"),
    (9, "September"),
    (10, "October"),
    (11, "November"),
    (12, "December"),
)

HOUR_CHOICES = [("", "")] + list(zip(*[range(0, 24, 1)] * 2))

MINUTES = range(0, 60)

MINUTE_CHOICES = [("", "")] + list(zip(MINUTES, ["%02d" % m for m in MINUTES]))

TIMEZONE_CHOICES = list(
    timezone_choices(pytz.country_names, pytz.country_timezones, itemgetter(1))
)

# Add UTC and MGT
TIMEZONE_CHOICES.insert(
    1,
    (
        _("Universal"),
        [
            ("UTC", _("Coordinated Universal Time")),
            ("GMT", _("Greenwich Mean Time")),
        ],
    ),
)


class SelectTimezoneForm(forms.Form):
    """
    Used by touchtechnology.common.context_processors.tz to generate a
    dropdown list of timezones to be used with the TimeZone middleware and
    set-timezone view.
    """

    timezone = forms.ChoiceField(
        choices=TIMEZONE_CHOICES, initial=timezone.get_current_timezone
    )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-07-06 11:59
from __future__ import unicode_literals

import touchtechnology.common.db.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('common', '0005_sitemapnode_populate_kwargs'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sitemapnode',
            name='enabled',
            field=touchtechnology.common.db.models.BooleanField(default=True, help_text="Set this to 'No' to disable this object and it's children on the site.", verbose_name='Enabled'),
        ),
        migrations.AlterField(
            model_name='sitemapnode',
            name='hidden_from_navigation',
            field=touchtechnology.common.db.models.BooleanField(default=False, help_text="When set to 'Yes' this object will still be available, but will not appear in menus.", verbose_name='Hide from menus'),
        ),
        migrations.AlterField(
            model_name='sitemapnode',
            name='hidden_from_robots',
            field=touchtechnology.common.db.models.BooleanField(default=False, help_text='Set this to <em>Yes</em> to prevent search engines from indexing this part of the site.<br><strong>Warning:</strong> this may affect your ranking in search engines.', verbose_name='Hide from spiders'),
        ),
        migrations.AlterField(
            model_name='sitemapnode',
            name='hidden_from_sitemap',
            field=touchtechnology.common.db.models.BooleanField(default=False, help_text="When set to 'Yes' this object will not be listed in the auto-generated sitemap.", verbose_name='Hide from sitemap'),
        ),
        migrations.AlterField(
            model_name='sitemapnode',
            name='last_modified',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name='sitemapnode',
            name='require_https',
            field=touchtechnology.common.db.models.BooleanField(default=False, help_text='Force this to be served via HTTPS.', verbose_name='HTTPS Required'),
        ),
        migrations.AlterField(
            model_name='sitemapnode',
            name='restrict_to_groups',
            field=models.ManyToManyField(blank=True, help_text='If you select one or more of these groups your visitors will need to be logged in and a member of an appropriate group to view this part of the site.', to='auth.Group', verbose_name='Restrict to Groups'),
        ),
        migrations.AlterField(
            model_name='sitemapnode',
            name='slug_locked',
            field=touchtechnology.common.db.models.BooleanField(default=False, verbose_name='Slug locked'),
        ),
    ]



from django.urls import include, path

from touchtechnology.common.sites import AccountsSite
from tournamentcontrol.competition.sites import competition

accounts = AccountsSite()

urlpatterns = [
    path("accounts/", accounts.urls),
    path("api/", include("tournamentcontrol.competition.rest.urls")),
    path("", competition.urls),
]



from django.db import migrations


def article_content_to_copy(apps, schema_editor):
    Article = apps.get_model("news", "Article")

    for article in Article.objects.prefetch_related("content"):
        article.copy = "\n".join(
            [
                "<!-- %(label)s -->\n%(copy)s\n<!-- /%(label)s -->\n" % data
                for data in article.content.values()
            ]
        )
        article.save()


class Migration(migrations.Migration):

    dependencies = [
        ("news", "0006_article_copy_field"),
    ]

    operations = [
        migrations.RunPython(
            article_content_to_copy, reverse_code=migrations.RunPython.noop
        )
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-07-06 11:59
from __future__ import unicode_literals

import django.db.models.deletion
import touchtechnology.common.db.models
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('competition', '0030_match_referees'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='seasonreferee',
            options={'ordering': ('season', 'person__last_name', 'person__first_name'), 'verbose_name': 'referee'},
        ),
        migrations.AlterField(
            model_name='club',
            name='primary',
            field=touchtechnology.common.db.models.ForeignKey(blank=True, help_text='Appears on the front-end with other club information.', null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to='competition.Person', verbose_name='Primary contact'),
        ),
        migrations.AlterField(
            model_name='club',
            name='slug_locked',
            field=touchtechnology.common.db.models.BooleanField(default=False, verbose_name='Slug locked'),
        ),
        migrations.AlterField(
            model_name='clubassociation',
            name='club',
            field=touchtechnology.common.db.models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='staff', to='competition.Club'),
        ),
        migrations.AlterField(
            model_name='clubassociation',
            name='person',
            field=touchtechnology.common.db.models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='competition.Person'),
        ),
        migrations.AlterField(
            model_name='clubrole',
            name='competition',
            field=touchtechnology.common.db.models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='club_roles', to='competition.Competition'),
        ),
        migrations.AlterField(
            model_name='competition',
            name='enabled',
            field=touchtechnology.common.db.models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name='competition',
            name='slug_locked',
            field=touchtechnology.common.db.models.BooleanField(default=False, verbose_name='Slug locked'),
        ),
        migrations.AlterField(
            model_name='division',
            name='draft',
            field=touchtechnology.common.db.models.BooleanField(default=False, help_text='Marking a division as draft will prevent matches from being visible in the front-end.'),
        ),
        migrations.AlterField(
            model_name='division',
            name='include_forfeits_in_played',
            field=touchtechnology.common.db.models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name='division',
            name='rank_division',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='competition.RankDivision'),
        ),
        migrations.AlterField(
            model_name='division',
            name='season',
            field=touchtechnology.common.db.models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='divisions', to='competition.Season'),
        ),
        migrations.AlterField(
            model_name='division',
            name='slug_locked',
            field=touchtechnology.common.db.models.BooleanField(default=False, verbose_name='Slug locked'),
        ),
        migrations.AlterField(
            model_name='divisionexclusiondate',
            name='date',
            field=models.DateField(),
        ),
        migrations.AlterField(
            model_name='drawformat',
            name='is_final',
            field=touchtechnology.common.db.models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='ground',
            name='venue',
            field=touchtechnology.common.db.models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='grounds', to='competition.Venue'),
        ),
        migrations.AlterField(
            model_name='ladderentry',
            name='opponent',
            field=touchtechnology.common.db.models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='competition.Team'),
        ),
        migrations.AlterField(
            model_name='laddersummary',
            name='team',
            field=touchtechnology.common.db.models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='ladder_summary', to='competition.Team'),
        ),
        migrations.AlterField(
            model_name='match',
            name='away_team',
            field=touchtechnology.common.db.models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='away_games', to='competition.Team'),
        ),
        migrations.AlterField(
            model_name='match',
            name='away_team_eval_related',
            field=touchtechnology.common.db.models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to='competition.Match'),
        ),
        migrations.AlterField(
            model_name='match',
            name='away_team_undecided',
            field=touchtechnology.common.db.models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='away_games', to='competition.UndecidedTeam'),
        ),
        migrations.AlterField(
            model_name='match',
            name='bye_processed',
            field=touchtechnology.common.db.models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='match',
            name='date',
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='match',
            name='datetime',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='match',
            name='forfeit_winner',
            field=touchtechnology.common.db.models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to='competition.Team'),
        ),
        migrations.AlterField(
            model_name='match',
            name='home_team',
            field=touchtechnology.common.db.models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='home_games', to='competition.Team'),
        ),
        migrations.AlterField(
            model_name='match',
            name='home_team_eval_related',
            field=touchtechnology.common.db.models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to='competition.Match'),
        ),
        migrations.AlterField(
            model_name='match',
            name='home_team_undecided',
            field=touchtechnology.common.db.models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='home_games', to='competition.UndecidedTeam'),
        ),
        migrations.AlterField(
            model_name='match',
            name='include_in_ladder',
            field=touchtechnology.common.db.models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name='match',
            name='is_bye',
            field=touchtechnology.common.db.models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='match',
            name='is_forfeit',
            field=touchtechnology.common.db.models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='match',
            name='is_washout',
            field=touchtechnology.common.db.models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='match',
            name='stage',
            field=touchtechnology.common.db.models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, related_name='matches', to='competition.Stage'),
        ),
        migrations.AlterField(
            model_name='match',
            name='stage_group',
            field=touchtechnology.common.db.models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='matches', to='competition.StageGroup', verbose_name='Pool'),
        ),
        migrations.AlterField(
            model_name='match',
            name='time',
            field=models.TimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='person',
            name='club',
            field=touchtechnology.common.db.models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='members', to='competition.Club'),
        ),
        migrations.AlterField(
            model_name='person',
            name='date_of_birth',
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='person',
            name='user',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='place',
            name='slug_locked',
            field=touchtechnology.common.db.models.BooleanField(default=False, verbose_name='Slug locked'),
        ),
        migrations.AlterField(
            model_name='rankdivision',
            name='enabled',
            field=touchtechnology.common.db.models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name='rankdivision',
            name='slug_locked',
            field=touchtechnology.common.db.models.BooleanField(default=False, verbose_name='Slug locked'),
        ),
        migrations.AlterField(
            model_name='season',
            name='competition',
            field=touchtechnology.common.db.models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='seasons', to='competition.Competition'),
        ),
        migrations.AlterField(
            model_name='season',
            name='complete',
            field=touchtechnology.common.db.models.BooleanField(default=False, help_text='Set to indicate this season is in the past. Optimises progression calculations.'),
        ),
        migrations.AlterField(
            model_name='season',
            name='enabled',
            field=touchtechnology.common.db.models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name='season',
            name='forfeit_notifications',
            field=touchtechnology.common.db.models.ManyToManyField(blank=True, help_text='When a team advises they are forfeiting, notify the opposition team plus these people.', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='season',
            name='mvp_results_public',
            field=models.DateTimeField(blank=True, help_text='The time when the results of the MVP voting will be made public on the website. Leave blank to show at all times.', null=True, verbose_name='MVP public at'),
        ),
        migrations.AlterField(
            model_name='season',
            name='slug_locked',
            field=touchtechnology.common.db.models.BooleanField(default=False, verbose_name='Slug locked'),
        ),
        migrations.AlterField(
            model_name='season',
            name='start_date',
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='season',
            name='statistics',
            field=touchtechnology.common.db.models.BooleanField(default=True, help_text='Set to No if you do not wish to keep scoring or most valuable player statistics.'),
        ),
        migrations.AlterField(
            model_name='seasonassociation',
            name='club',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='officials', to='competition.Club'),
        ),
        migrations.AlterField(
            model_name='seasonassociation',
            name='person',
            field=touchtechnology.common.db.models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='competition.Person'),
        ),
        migrations.AlterField(
            model_name='seasonassociation',
            name='season',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='officials', to='competition.Season'),
        ),
        migrations.AlterField(
            model_name='seasonexclusiondate',
            name='date',
            field=models.DateField(),
        ),
        migrations.AlterField(
            model_name='seasonmatchtime',
            name='end_date',
            field=models.DateField(blank=True, null=True, verbose_name='Until'),
        ),
        migrations.AlterField(
            model_name='seasonmatchtime',
            name='start',
            field=models.TimeField(),
        ),
        migrations.AlterField(
            model_name='seasonmatchtime',
            name='start_date',
            field=models.DateField(blank=True, null=True, verbose_name='From'),
        ),
        migrations.AlterField(
            model_name='simplescorematchstatistic',
            name='match',
            field=touchtechnology.common.db.models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='statistics', to='competition.Match'),
        ),
        migrations.AlterField(
            model_name='simplescorematchstatistic',
            name='player',
            field=touchtechnology.common.db.models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='statistics', to='competition.Person'),
        ),
        migrations.AlterField(
            model_name='stage',
            name='carry_ladder',
            field=touchtechnology.common.db.models.BooleanField(default=False, help_text='Set this to <b>Yes</b> if this stage should carry over values from the previous stage.', verbose_name='Carry over points'),
        ),
        migrations.AlterField(
            model_name='stage',
            name='division',
            field=touchtechnology.common.db.models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='stages', to='competition.Division'),
        ),
        migrations.AlterField(
            model_name='stage',
            name='follows',
            field=touchtechnology.common.db.models.ForeignKey(blank=True, help_text='When progressing teams into this stage, which earlier stage should be used for determining positions.<br>Default is the immediately preceeding stage.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='preceeds', to='competition.Stage'),
        ),
        migrations.AlterField(
            model_name='stage',
            name='keep_ladder',
            field=touchtechnology.common.db.models.BooleanField(default=True, help_text='Set this to <b>No</b> if this stage does not need to keep a competition ladder.<br>Usually set to No for a Final Series or a Knockout stage.', verbose_name='Keep a ladder'),
        ),
        migrations.AlterField(
            model_name='stage',
            name='keep_mvp',
            field=touchtechnology.common.db.models.BooleanField(default=True, help_text='Set this to <b>No</b> if this stage does not need to keep track of MVP points.<br>Usually set to No for a Final Series.', verbose_name='Keep MVP stats'),
        ),
        migrations.AlterField(
            model_name='stage',
            name='scale_group_points',
            field=touchtechnology.common.db.models.BooleanField(default=False, help_text='In stages with multiple pools, adjust points in the smaller groups to compensate for the reduced opportunity to score points.<br>You <strong>should</strong> also set 0 points for Bye matches.'),
        ),
        migrations.AlterField(
            model_name='stage',
            name='slug_locked',
            field=touchtechnology.common.db.models.BooleanField(default=False, verbose_name='Slug locked'),
        ),
        migrations.AlterField(
            model_name='stagegroup',
            name='carry_ladder',
            field=touchtechnology.common.db.models.BooleanField(default=False, help_text='Set this to <b>Yes</b> if the ladder for this pool should carry over values from the previous stage.<br>Will only apply for matches played against teams that are now in this group.', verbose_name='Carry over points'),
        ),
        migrations.AlterField(
            model_name='stagegroup',
            name='slug_locked',
            field=touchtechnology.common.db.models.BooleanField(default=False, verbose_name='Slug locked'),
        ),
        migrations.AlterField(
            model_name='stagegroup',
            name='stage',
            field=touchtechnology.common.db.models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='pools', to='competition.Stage'),
        ),
        migrations.AlterField(
            model_name='team',
            name='division',
            field=touchtechnology.common.db.models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='teams', to='competition.Division'),
        ),
        migrations.AlterField(
            model_name='team',
            name='names_locked',
            field=touchtechnology.common.db.models.BooleanField(default=False, help_text='When the team name is locked, the team manager will not be able to change their team name.<br>As a tournament manager you can always change the names.'),
        ),
        migrations.AlterField(
            model_name='team',
            name='rank_division',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='competition.RankDivision'),
        ),
        migrations.AlterField(
            model_name='team',
            name='slug_locked',
            field=touchtechnology.common.db.models.BooleanField(default=False, verbose_name='Slug locked'),
        ),
        migrations.AlterField(
            model_name='team',
            name='timeslots_after',
            field=models.TimeField(blank=True, help_text='Specify the earliest time that this team can play. Leave blank for no preference.', null=True, verbose_name='Start after'),
        ),
        migrations.AlterField(
            model_name='team',
            name='timeslots_before',
            field=models.TimeField(blank=True, help_text='Specify the latest time that this team can play. Leave blank for no preference.', null=True, verbose_name='Start before'),
        ),
        migrations.AlterField(
            model_name='teamassociation',
            name='is_player',
            field=touchtechnology.common.db.models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name='teamassociation',
            name='person',
            field=touchtechnology.common.db.models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='competition.Person'),
        ),
        migrations.AlterField(
            model_name='teamassociation',
            name='team',
            field=touchtechnology.common.db.models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='people', to='competition.Team'),
        ),
        migrations.AlterField(
            model_name='teamrole',
            name='competition',
            field=touchtechnology.common.db.models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='team_roles', to='competition.Competition'),
        ),
        migrations.AlterField(
            model_name='undecidedteam',
            name='stage',
            field=touchtechnology.common.db.models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='undecided_teams', to='competition.Stage'),
        ),
        migrations.AlterField(
            model_name='venue',
            name='season',
            field=touchtechnology.common.db.models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='venues', to='competition.Season'),
        ),
    ]



# encoding=utf8

from __future__ import unicode_literals

import re
from urllib.parse import urlparse

from django.core import mail
from test_plus import TestCase
from touchtechnology.common.tests import factories

# Based on regular expression found at https://gist.github.com/gruber/8891611
url_re = re.compile(
    r"""# noqa
(?xi)
\b
(                                             # Capture 1: entire matched URL
  (?:
    https?:                                   # URL protocol and colon
    (?:
      /{1,3}                                  # 1-3 slashes
      |                                       #   or
      [a-z0-9%]                               # Single letter or digit or '%'
                                              # (Trying not to match e.g. "URI::Escape")
    )
    |                                         #   or
                                              # looks like domain name followed by a slash:
    [a-z0-9.\-]+[.]
    (?:com|net|org|edu|gov|mil|aero|asia|biz|cat|coop|info|int|jobs|mobi|museum|name|post|pro|tel|travel|xxx|ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj| Ja|sk|sl|sm|sn|so|sr|ss|st|su|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|yu|za|zm|zw)
    /
  )
  (?:                                         # One or more:
    [^\s()<>{}\[\]]+                          # Run of non-space, non-()<>{}[]
    |                                         #   or
    \([^\s()]*?\([^\s()]+\)[^\s()]*?\)        # balanced parens, one level deep: (â€¦(â€¦)â€¦)
    |
    \([^\s]+?\)                               # balanced parens, non-recursive: (â€¦)
  )+
  (?:                                         # End with:
    \([^\s()]*?\([^\s()]+\)[^\s()]*?\)        # balanced parens, one level deep: (â€¦(â€¦)â€¦)
    |
    \([^\s]+?\)                               # balanced parens, non-recursive: (â€¦)
    |                                         #   or
    [^\s`!()\[\]{};:'".,<>?Â«Â»â€œâ€â€˜â€™] # not a space or one of these punct chars
  )
#  |                                           # OR, the following to match naked domains:
#  (?:
#      (?<!@)                                  # not preceded by a @, avoid matching foo@_gmail.com_
#    [a-z0-9]+
#    (?:[.\-][a-z0-9]+)*
#    [.]
#    (?:com|net|org|edu|gov|mil|aero|asia|biz|cat|coop|info|int|jobs|mobi|museum|name|post|pro|tel|travel|xxx|ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj| Ja|sk|sl|sm|sn|so|sr|ss|st|su|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|yu|za|zm|zw)
#    \b
#    /?
#    (?!@)                                     # not succeeded by a @, avoid matching "foo.na" in "foo.na@example.com"
#  )
)
""",
    re.VERBOSE,
)


class PasswordResetTests(TestCase):
    def setUp(self):
        self.staff = factories.UserFactory.create(is_staff=True, is_superuser=True)
        self.regular = factories.UserFactory.create()

    def test_login(self):
        self.assertGoodView("accounts:login")

    def test_logout(self):
        self.assertGoodView("accounts:logout")

    def test_password_change(self):
        with self.login(self.regular):
            self.assertGoodView("accounts:password_change")

    def test_password_change_done(self):
        with self.login(self.regular):
            self.assertGoodView("accounts:password_change_done")

    def test_password_reset_complete(self):
        self.assertGoodView("accounts:password_reset_complete")

    def test_password_reset_confirm(self):
        self.assertGoodView("accounts:password_reset_confirm", uidb64="z", token="1-a")

    def test_password_reset_GET(self):
        self.assertGoodView("accounts:password_reset")

    def test_password_reset_POST(self):
        data = {
            "email": self.staff.email,
        }
        self.post("accounts:password_reset", data=data)

        # check that the form submission was successful and
        # redirection to the "done" page has occured.
        redirect_to = self.reverse("accounts:password_reset_done")
        self.assertRedirects(self.last_response, redirect_to)

        # check that an email has been delivered
        self.assertEqual(len(mail.outbox), 1)

        # check that the email body is correct
        text_re = re.compile(
            """Just in case you've forgotten, your username is ".+"."""
        )
        self.assertRegex(mail.outbox[0].body, text_re)

        # check that the URL in the email body works
        link = url_re.findall(mail.outbox[0].body)[0]
        self.get(urlparse(link).path, follow=True)
        self.assertResponseContains("<h1>Enter new password</h1>")



from __future__ import unicode_literals

import mock
from test_plus import TestCase
from touchtechnology.common.forms.fields import boolean_choice_field_coerce
from touchtechnology.common.forms.tz import timezone_choice
from touchtechnology.common.utils import get_base_url, get_mod_func


class BooleanChoiceFieldCoerce(TestCase):

    def test_true_boolean(self):
        result = boolean_choice_field_coerce(True)
        self.assertTrue(result)

    def test_false_boolean(self):
        result = boolean_choice_field_coerce(False)
        self.assertFalse(result)

    def test_true_integer(self):
        result = boolean_choice_field_coerce('1')
        self.assertTrue(result)

    def test_false_integer(self):
        result = boolean_choice_field_coerce('0')
        self.assertFalse(result)

    def test_true_string(self):
        result = boolean_choice_field_coerce(1)
        self.assertTrue(result)

    def test_false_string(self):
        result = boolean_choice_field_coerce(0)
        self.assertFalse(result)


class GetModFunc(TestCase):

    def test_get_mod_func(self):
        mod_name, func_name = get_mod_func(
            'django.contrib.auth.tokens.default_token_generator')
        self.assertEqual(mod_name, 'django.contrib.auth.tokens')
        self.assertEqual(func_name, 'default_token_generator')

    def test_no_dot(self):
        mod_name, func_name = get_mod_func('django')
        self.assertEqual(mod_name, 'django')
        self.assertEqual(func_name, '')


class TimeZoneChoice(TestCase):

    def test_one_part(self):
        self.assertEqual(
            timezone_choice('UTC'),
            ('UTC', 'UTC'))

    def test_two_parts(self):
        self.assertEqual(
            timezone_choice('Australia/Sydney'),
            ('Australia/Sydney', 'Sydney'))

    def test_three_parts(self):
        self.assertEqual(
            timezone_choice('America/Indiana/Indianapolis'),
            ('America/Indiana/Indianapolis', 'Indianapolis (Indiana)'))


class GetBaseUrl(TestCase):

    @mock.patch('touchtechnology.common.utils.connection')
    def test_tenant_no_prepend_www(self, connection):
        connection.tenant = mock.Mock(['domain_url'])
        connection.tenant.domain_url = 'example.com'
        self.assertEqual('http://example.com/', get_base_url())

    @mock.patch('touchtechnology.common.utils.connection')
    def test_tenant_prepend_www_true(self, connection):
        connection.tenant = mock.Mock(['domain_url', 'prepend_www'])
        connection.tenant.domain_url = 'example.com'
        connection.tenant.prepend_www = True
        self.assertEqual('http://www.example.com/', get_base_url())

    @mock.patch('touchtechnology.common.utils.connection')
    def test_tenant_prepend_www_false(self, connection):
        connection.tenant = mock.Mock(['domain_url', 'prepend_www'])
        connection.tenant.domain_url = 'example.com'
        connection.tenant.prepend_www = False
        self.assertEqual('http://example.com/', get_base_url())

    def test_site(self):
        self.assertEqual('http://example.com/', get_base_url())



from __future__ import unicode_literals

import base64
import collections
import functools
import logging
import operator

from django.apps import apps
from django.conf import settings
from django.contrib import messages
from django.db import models
from django.db.models import Case, F, Q, Sum, When
from django.forms.models import _get_foreign_key
from django.http import Http404, HttpResponse, HttpResponseGone
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse
from django.urls import include, path, re_path, reverse
from django.utils import timezone
from django.utils.translation import gettext_lazy as _, ngettext

from touchtechnology.admin.base import AdminComponent
from touchtechnology.common.decorators import (
    csrf_exempt_m,
    staff_login_required_m,
)
from touchtechnology.common.prince import prince
from tournamentcontrol.competition.dashboard import (
    BasicResultWidget,
    DetailResultWidget,
    MostValuableWidget,
    ProgressStageWidget,
    ScoresheetWidget,
)
from tournamentcontrol.competition.decorators import (
    competition_by_pk_m,
    registration,
)
from tournamentcontrol.competition.forms import (
    ClubAssociationForm,
    ClubRoleForm,
    CompetitionForm,
    DivisionForm,
    DrawFormatForm,
    DrawGenerationFormSet,
    DrawGenerationMatchFormSet,
    GroundForm,
    MatchEditForm,
    MatchScheduleFormSet,
    MatchWashoutFormSet,
    PersonEditForm,
    PersonMergeForm,
    ProgressMatchesFormSet,
    ProgressTeamsFormSet,
    RescheduleDateFormSet,
    SeasonAssociationFormSet,
    SeasonForm,
    SeasonMatchTimeFormSet,
    StageForm,
    StageGroupForm,
    TeamAssociationForm,
    TeamAssociationFormSet,
    TeamForm,
    TeamRoleForm,
    UndecidedTeamForm,
    VenueForm,
)
from tournamentcontrol.competition.models import (
    Club,
    ClubAssociation,
    ClubRole,
    Competition,
    Division,
    DivisionExclusionDate,
    DrawFormat,
    Ground,
    LadderEntry,
    LadderSummary,
    Match,
    MatchScoreSheet,
    Person,
    Season,
    SeasonAssociation,
    SeasonExclusionDate,
    SeasonMatchTime,
    SeasonReferee,
    SimpleScoreMatchStatistic,
    Stage,
    StageGroup,
    Team,
    TeamAssociation,
    TeamRole,
    UndecidedTeam,
    Venue,
)
from tournamentcontrol.competition.sites import CompetitionAdminMixin
from tournamentcontrol.competition.tasks import generate_pdf_scorecards
from tournamentcontrol.competition.utils import (
    generate_fixture_grid,
    generate_scorecards,
    legitimate_bye_match,
    match_unplayed,
    team_needs_progressing,
)
from tournamentcontrol.competition.wizards import DrawGenerationWizard

SCORECARD_PDF_WAIT = getattr(settings, "TOURNAMENTCONTROL_SCORECARD_PDF_WAIT", 5)

log = logging.getLogger(__name__)


def next_related_factory(model, parent=None, fk_name=None):
    kw = {}
    if parent is not None:
        if fk_name is None:
            fk = _get_foreign_key(parent.__class__, model)
            kw = {fk.name: parent}
        else:
            kw = {fk_name: parent}
    last = (
        model.objects.filter(**kw).aggregate(order=models.Max("order")).get("order")
        or 0
    )
    return model(order=last + 1, **kw)


class CompetitionAdminComponent(CompetitionAdminMixin, AdminComponent):
    verbose_name = _("Tournament Control")
    unprotected = False

    def __init__(self, app, name="fixja", app_name="competition"):
        super(CompetitionAdminComponent, self).__init__(app, name, app_name)

    def dropdowns(self):
        dl = (
            (_("Competitions"), self.reverse("competition:list"), "trophy"),
            (_("Clubs"), self.reverse("club:list"), "shield"),
            (_("Draw Formats"), self.reverse("format:list"), "wrench"),
            (_("Reports"), self.reverse("scorecard-report"), "book"),
        )
        return dl

    @property
    def widgets(self):
        widget_classes = (
            BasicResultWidget,
            ProgressStageWidget,
            ScoresheetWidget,
            DetailResultWidget,
            MostValuableWidget,
        )
        return widget_classes

    def render(self, request, templates, context, *args, **kwargs):
        context.update(
            {
                "base": "tournamentcontrol/rego/admin/base.html",
            }
        )
        return super(CompetitionAdminComponent, self).render(
            request, templates, context, *args, **kwargs
        )

    def get_urls(self):
        stub_urls = (
            # Used to prevent template rendering failures; all uses *should* be marked
            # with a FIXME comment.
            [
                path("add/", self.index, name="add"),
            ],
            self.app_name,
        )

        matchscoresheet_urls = (
            [
                path("add/", self.edit_matchscoresheet, name="add"),
                path("<int:pk>/", self.edit_matchscoresheet, name="edit"),
                path("<int:pk>/delete/", self.delete_matchscoresheet, name="delete"),
            ],
            self.app_name,
        )

        match_urls = (
            [
                path("add/", self.edit_match, name="add"),
                path("<int:match_id>/", self.edit_match, name="edit"),
                path("<int:match_id>/delete/", self.delete_match, name="delete"),
                path("<int:match_id>/detail/", self.edit_match_detail, name="detail"),
                path(
                    "<int:match_id>/scoresheet/",
                    include(matchscoresheet_urls, namespace="matchscoresheet"),
                ),
                # FIXME - these just prevent template rendering failures
                path(
                    "<int:match_id>/ladder/",
                    include(stub_urls, namespace="ladderentry"),
                ),
                path(
                    "<int:match_id>/statistic/",
                    include(stub_urls, namespace="simplescorematchstatistic"),
                ),
            ],
            self.app_name,
        )

        role_patterns = (
            [
                path("add/", self.edit_role, name="add"),
                path("<int:pk>/", self.edit_role, name="edit"),
                path("<int:pk>/delete/", self.delete_role, name="delete"),
            ],
            self.app_name,
        )

        timeslot_urls = (
            [
                path("add/", self.edit_timeslot, name="add"),
                path("<int:pk>/", self.edit_timeslot, name="edit"),
                path("<int:pk>/delete/", self.delete_timeslot, name="delete"),
            ],
            self.app_name,
        )

        referees_urls = (
            [
                path("add/", self.edit_seasonreferee, name="add"),
                path("<int:pk>/", self.edit_seasonreferee, name="edit"),
                path("<int:pk>/delete/", self.delete_seasonreferee, name="delete"),
            ],
            self.app_name,
        )

        ground_urls = (
            [
                path(r"add/", self.edit_ground, name="add"),
                path("<int:ground_id>/", self.edit_ground, name="edit"),
                path("<int:ground_id>/delete/", self.delete_ground, name="delete"),
            ],
            self.app_name,
        )

        venue_urls = (
            [
                path("add/", self.edit_venue, name="add"),
                path("<int:venue_id>/", self.edit_venue, name="edit"),
                path("<int:venue_id>/delete/", self.delete_venue, name="delete"),
                path(
                    "<int:venue_id>/ground/", include(ground_urls, namespace="ground")
                ),
            ],
            self.app_name,
        )

        seasonexclusion_urls = (
            [
                path("add/", self.edit_seasonexclusiondate, name="add"),
                path("<int:pk>/", self.edit_seasonexclusiondate, name="edit"),
                path(
                    "<int:pk>/delete/", self.delete_seasonexclusiondate, name="delete"
                ),
            ],
            self.app_name,
        )

        divisionexclusion_urls = (
            [
                path("add/", self.edit_divisionexclusiondate, name="add"),
                path("<int:pk>/", self.edit_divisionexclusiondate, name="edit"),
                path(
                    "<int:pk>/delete/", self.delete_divisionexclusiondate, name="delete"
                ),
            ],
            self.app_name,
        )

        draw_urls = (
            [
                # path("", self.list_draw, name="list"),
                path("build/", self.generate_draw, name="build"),
                path("undo/", self.undo_draw, name="undo"),
                path("progress/", self.progress_teams, name="progress"),
            ],
            self.app_name,
        )

        undecidedteam_urls = (
            [
                path(r"add/", self.edit_team, name="add"),
                path(r"<int:team_id>/", self.edit_team, name="edit"),
                path(r"<int:team_id>/delete/", self.delete_team, name="delete"),
                path(r"<int:team_id>/match/", include(match_urls, namespace="match")),
            ],
            self.app_name,
        )

        stagegroup_urls = (
            [
                # path("", self.list_pools, name="list"),
                path("add/", self.edit_pool, name="add"),
                path("<int:pool_id>/", self.edit_pool, name="edit"),
                path("<int:pool_id>/delete/", self.delete_pool, name="delete"),
                path("<int:pool_id>/match/", include(match_urls, namespace="match")),
            ],
            self.app_name,
        )

        stage_urls = (
            [
                path("add/", self.edit_stage, name="add"),
                path("<int:stage_id>/", self.edit_stage, name="edit"),
                path("<int:stage_id>/delete/", self.delete_stage, name="delete"),
                path("<int:stage_id>/match/", include(match_urls, namespace="match")),
                path("<int:stage_id>/draw/", include(draw_urls, namespace="draw")),
                path(
                    "<int:stage_id>/team/",
                    include(undecidedteam_urls, namespace="undecidedteam"),
                ),
                path(
                    "<int:stage_id>/scorecards.<mode>",
                    self.scorecards,
                    name="scorecards",
                ),
                path(
                    "<int:stage_id>/pool/",
                    include(stagegroup_urls, namespace="stagegroup"),
                ),
            ],
            self.app_name,
        )

        teamassociation_urls = (
            [
                path("add/", self.edit_teamassociation, name="add"),
                path("<int:pk>/", self.edit_teamassociation, name="edit"),
                path("<int:pk>/delete/", self.delete_teamassociation, name="delete"),
            ],
            self.app_name,
        )

        team_urls = (
            [
                path("add/", self.edit_team, name="add"),
                path("<int:team_id>/", self.edit_team, name="edit"),
                path("<int:team_id>/delete/", self.delete_team, name="delete"),
                path("<int:team_id>/permissions/", self.perms_team, name="perms"),
                path("<int:team_id>/match/", include(match_urls, namespace="match")),
                path(
                    "<int:team_id>/association/",
                    include(teamassociation_urls, namespace="teamassociation"),
                ),
            ],
            self.app_name,
        )

        division_urls = (
            [
                path(r"add/", self.edit_division, name="add"),
                path("<int:division_id>/", self.edit_division, name="edit"),
                path("<int:division_id>/delete/", self.delete_division, name="delete"),
                path(
                    "<int:division_id>/stage/", include(stage_urls, namespace="stage")
                ),
                path("<int:division_id>/teams/", include(team_urls, namespace="team")),
                path(
                    "<int:division_id>/scorers/",
                    self.highest_point_scorer,
                    name="scorers",
                ),
                path(
                    "<int:division_id>/exclusion/",
                    include(divisionexclusion_urls, namespace="divisionexclusiondate"),
                ),
            ],
            self.app_name,
        )

        season_urls = (
            [
                path("add/", self.edit_season, name="add"),
                path("<int:season_id>/", self.edit_season, name="edit"),
                path("<int:season_id>/delete/", self.delete_season, name="delete"),
                path(
                    "<int:season_id>/reschedule/",
                    self.match_reschedule,
                    name="reschedule",
                ),
                path(
                    "<int:season_id>/timeslot/",
                    include(timeslot_urls, namespace="seasonmatchtime"),
                ),
                path(
                    "<int:season_id>/referees/",
                    include(referees_urls, namespace="seasonreferee"),
                ),
                path("<int:season_id>/permission/", self.perms_season, name="perms"),
                path("<int:season_id>/venue/", include(venue_urls, namespace="venue")),
                path(
                    "<int:season_id>/exclusion/",
                    include(seasonexclusion_urls, namespace="seasonexclusiondate"),
                ),
                path(
                    "<int:season_id>/division/",
                    include(division_urls, namespace="division"),
                ),
                path(
                    "<int:season_id>/scorecards/<result_id>.pdf",
                    self.scorecards_async,
                    name="scorecards-async",
                ),
                # previously was not in this namespace
                path("<int:season_id>/report.html", self.season_report, name="report"),
                path(
                    "<int:season_id>/summary.html", self.season_summary, name="summary"
                ),
                path(
                    "<int:season_id>/matches/grid.<mode>",
                    self.season_grid,
                    name="match-grid",
                ),
            ],
            self.app_name,
        )

        competition_urls = (
            [
                path("", self.list_competitions, name="list"),  # KEEP
                path("add/", self.edit_competition, name="add"),
                path("<int:competition_id>/", self.edit_competition, name="edit"),
                path(
                    "<int:competition_id>/delete/",
                    self.delete_competition,
                    name="delete",
                ),
                path(
                    "<int:competition_id>/permission/",
                    self.perms_competition,
                    name="perms",
                ),
                path(
                    "<int:competition_id>/seasons/",
                    include(season_urls, namespace="season"),
                ),
                path(
                    "<int:competition_id>/role-club/",
                    include(role_patterns, namespace="clubrole"),
                    {"cls": "club"},
                ),
                path(
                    "<int:competition_id>/role-team/",
                    include(role_patterns, namespace="teamrole"),
                    {"cls": "team"},
                ),
            ],
            self.app_name,
        )

        person_urls = (
            [
                path("add/", self.edit_person, name="add"),
                path("<uuid:person_id>/", self.edit_person, name="edit"),
                path("<uuid:person_id>/delete/", self.delete_person, name="delete"),
                path("<uuid:person_id>/merge/", self.merge_person, name="merge"),
            ],
            self.app_name,
        )

        clubassociation_urls = (
            [
                path("add/", self.edit_clubassociation, name="add"),
                path(
                    "<int:clubassociation_id>/", self.edit_clubassociation, name="edit"
                ),
            ],
            self.app_name,
        )

        club_urls = (
            [
                path("", self.list_clubs, name="list"),
                path("add/", self.edit_club, name="add"),
                path("<int:club_id>/", self.edit_club, name="edit"),
                path("<int:club_id>/delete/", self.delete_club, name="delete"),
                path("<int:club_id>/person/", include(person_urls, namespace="person")),
                # From RegistrationBase
                path(
                    "<int:club_id>/officials/<season_id>/",
                    self.officials,
                    name="officials",
                ),
                path(
                    "<int:club_id>/<season_id>/registration.<mode>",
                    self.registration_form,
                    name="registration-form",
                ),
                # FIXME should have a 'season' namespace
                path(
                    "<int:club_id>/<season_id>/team/",
                    include(
                        (
                            [
                                # path("add/", self.edit_team_members, name="add"),
                                path(
                                    "<int:team_id>/",
                                    self.edit_team_members,
                                    name="edit",
                                ),
                            ],
                            self.app_name,
                        ),
                        namespace="team",
                    ),
                ),
                path(
                    "<int:club_id>/clubassociation/",
                    include(clubassociation_urls, namespace="clubassociation"),
                ),
            ],
            self.app_name,
        )

        drawformat_urls = (
            [
                path("", self.list_drawformat, name="list"),
                path("add/", self.edit_drawformat, name="add"),
                path("<int:pk>/", self.edit_drawformat, name="edit"),
                path("<int:pk>/delete/", self.delete_drawformat, name="delete"),
            ],
            self.app_name,
        )

        matchdatetime_urls = [
            path("results/", self.match_results, name="match-results"),
            path(
                "results:<int:division_id>/", self.match_results, name="match-results"
            ),
            path("washout/", self.match_washout, name="match-washout"),
            path("schedule/", self.match_schedule, name="match-schedule"),
            path(
                "schedule/<int:division_id>/",
                self.match_schedule,
                name="match-schedule",
            ),
            path(
                "schedule/<int:division_id>/<int:stage_id>/",
                self.match_schedule,
                name="match-schedule",
            ),
            path(
                "schedule/<int:division_id>/<int:stage_id>/<int:round>/",
                self.match_schedule,
                name="match-schedule",
            ),
            path("scorecards.<mode>", self.scorecards, name="scorecards"),
            path("runsheet.html", self.day_runsheet, name="match-runsheet"),
            path("grid.<mode>", self.day_grid, name="match-grid"),
        ]

        urlpatterns = [
            path("", self.index, name="index"),
            path("competition/", include(competition_urls, namespace="competition")),
            path(
                "competition/<int:competition_id>/seasons/<int:season_id>/<datestr>/",
                include(matchdatetime_urls),
            ),
            path(
                "<int:competition_id>/seasons/<int:season_id>:<timestr>/<datestr>/",
                include(matchdatetime_urls),
            ),
            path("club/", include(club_urls, namespace="club")),
            path("scorecards/", self.scorecard_report, name="scorecard-report"),
            path("draw-format/", include(drawformat_urls, namespace="format")),
            re_path(
                r"^reorder/(?P<model>[^/:]+)(?::(?P<parent>[^/]+))?/(?P<pk>\d+)/(?P<direction>\w+)/$",
                self.reorder,
                name="reorder",
            ),
        ]

        return urlpatterns

    @staff_login_required_m
    def reorder(self, request, model, pk, direction, parent=None, **kwargs):
        # Determine where to redirect back to following the reordering
        redirect_to = request.META.get("HTTP_REFERER", None)
        if redirect_to is None:
            raise Http404

        # Reconstruct the object from the keyword arguments
        cls = apps.get_model("competition", model)
        obj = cls.objects.get(pk=pk)

        # Determine the neighbouring order values based on direction
        if direction == "up":
            orders = obj.order, obj.order - 1
        elif direction == "down":
            orders = obj.order, obj.order + 1
        else:
            raise Http404

        # Get our OrderedSitemapNode queryset for the given model
        queryset = cls.objects.filter(order__in=orders)
        if parent is not None:
            queryset = queryset.filter(**{str(parent): getattr(obj, parent)})

        # We can't reorder the first item up or the last item down; throw
        # a 404 when trying to do so.
        if queryset.count() != 2:
            raise Http404

        # Swap the order value of the two neighbours and write to db.
        # 1) Hang onto the original ordering of a and b
        a, b = queryset
        ao, bo = a.order, b.order

        # 2) Set a.order to 0 to allow b.order to assume that value
        a.order = 0
        a.save()

        # 3) Assign the new value to b.order
        b.order = ao
        b.save()

        # 4) Assign the new value to a.order
        a.order = bo
        a.save()

        fmt = _(
            'The %(model)s "%(title)s" has been ' "reordered %(direction)s the list."
        )
        messages.info(
            request,
            fmt
            % dict(model=cls._meta.verbose_name, title=obj.title, direction=direction),
        )

        return self.redirect(redirect_to)

    @staff_login_required_m
    def index(self, request, **kwargs):
        messages.warning(
            request,
            _(
                "We're moving a few things around, please "
                "update any bookmarks you may find that "
                "have broken."
            ),
        )
        return self.redirect(self.reverse("competition:list"))

    @competition_by_pk_m
    @staff_login_required_m
    def edit_role(self, request, cls, pk=None, **extra_context):
        if cls == "club":
            model, form_class = ClubRole, ClubRoleForm
        elif cls == "team":
            model, form_class = TeamRole, TeamRoleForm
        else:
            raise Http404  # noqa

        competition = extra_context.get("competition")

        if pk is None:
            instance = model(competition=competition)
        else:
            instance = get_object_or_404(model, pk=pk, competition=competition)

        return self.generic_edit(
            request,
            model,
            instance=instance,
            form_class=form_class,
            permission_required=True,
            post_save_redirect=self.redirect(competition.urls["edit"]),
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def delete_role(self, request, cls, pk, extra_context, **kwargs):
        if cls == "club":
            model = ClubRole
        elif cls == "team":
            model = TeamRole
        else:
            raise Http404
        competition = extra_context.get("competition")
        queryset = model.objects.filter(competition=competition)
        post_delete_redirect = self.redirect(
            competition.urls["edit"] + "#%s_roles-tab" % cls
        )
        return self.generic_delete(
            request,
            queryset,
            pk=pk,
            permission_required=True,
            post_delete_redirect=post_delete_redirect,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def edit_matchscoresheet(self, request, pk=None, **extra_context):
        match = extra_context.get("match")

        if pk is None:
            instance = MatchScoreSheet(match=match)
        else:
            instance = get_object_or_404(MatchScoreSheet, pk=pk, match=match)

        return self.generic_edit(
            request,
            MatchScoreSheet,
            instance=instance,
            form_fields=("image",),
            permission_required=True,
            post_save_redirect=self.redirect(match.urls["edit"]),
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def delete_matchscoresheet(self, request, pk, **extra_context):
        return self.generic_delete(
            request, MatchScoreSheet, pk=pk, permission_required=True
        )

    @staff_login_required_m
    def list_drawformat(self, request, **extra_context):
        return self.generic_list(
            request,
            DrawFormat,
            paginate_by=0,
            permission_required=True,
            extra_context=extra_context,
        )

    @staff_login_required_m
    def edit_drawformat(self, request, pk=None, **extra_context):
        post_save_redirect = self.redirect(self.reverse("format:list"))
        return self.generic_edit(
            request,
            DrawFormat,
            pk=pk,
            form_class=DrawFormatForm,
            permission_required=True,
            post_save_redirect=post_save_redirect,
            extra_context=extra_context,
        )

    @staff_login_required_m
    def delete_drawformat(self, request, pk, **kwargs):
        return self.generic_delete(request, DrawFormat, pk=pk, permission_required=True)

    @staff_login_required_m
    def list_competitions(self, request, **extra_context):
        return self.generic_list(
            request,
            Competition,
            paginate_by=0,
            permission_required=True,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def perms_competition(self, request, competition, **kwargs):
        return self.generic_permissions(
            request, Competition, instance=competition, **kwargs
        )

    @competition_by_pk_m
    @staff_login_required_m
    def edit_competition(self, request, extra_context, competition=None, **kwargs):
        if competition is None:
            competition = next_related_factory(Competition)
        return self.generic_edit(
            request,
            Competition,
            instance=competition,
            form_class=CompetitionForm,
            form_kwargs={"user": request.user},
            permission_required=True,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def delete_competition(self, request, competition, extra_context, **kwargs):
        return self.generic_delete(
            request,
            Competition,
            pk=competition.pk,
            permission_required=True,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def edit_season(self, request, competition, extra_context, season=None, **kwargs):
        if season is None:
            season = next_related_factory(Season, competition)

        dates = season.matches.dates("date", "day")
        extra_context.setdefault("dates", dates)

        return self.generic_edit(
            request,
            Season,
            instance=season,
            related=(
                "exclusions",
                "divisions",
                "venues",
                "timeslots",
                "referees",
            ),
            form_class=SeasonForm,
            form_kwargs={"user": request.user},
            post_save_redirect=self.redirect(competition.urls["edit"]),
            permission_required=True,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def delete_season(self, request, extra_context, season, **kwargs):
        return self.generic_delete(
            request,
            Season,
            pk=season.pk,
            permission_required=True,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def perms_season(self, request, season, **kwargs):
        post_save_redirect = self.redirect(season.competition.urls["edit"])
        return self.generic_permissions(
            request,
            Season,
            instance=season,
            post_save_redirect=post_save_redirect,
            **kwargs,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def season_report(self, request, season, extra_context, **kwargs):
        teams = (
            Team.objects.filter(division__season=season)
            .order_by("club", "division", "order")
            .select_related("club")
        )
        return self.generic_list(
            request,
            teams,
            templates=self.template_path("season_report.html"),
            paginate_by=0,
            permission_required=False,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def season_summary(self, request, competition, season, extra_context, **kwargs):
        teams = (
            Team.objects.select_related("club", "division")
            .filter(division__season=season)
            .order_by("club__title", "division__order")
            .annotate(
                player_count=Sum(
                    Case(
                        When(people__is_player=True, then=1),
                        output_field=models.IntegerField(),
                    )
                ),
                non_player_count=Sum(
                    Case(
                        When(people__is_player=False, then=1),
                        output_field=models.IntegerField(),
                    )
                ),
            )
        )

        context = {
            "teams": teams,
            "season": season,
            "competition": competition,
        }
        context.update(extra_context)

        templates = self.template_path("season_summary.html")
        return self.render(request, templates, context)

    @competition_by_pk_m
    @staff_login_required_m
    def edit_timeslot(self, request, extra_context, season, pk=None, **kwargs):
        instance = SeasonMatchTime(season=season) if pk is None else None
        return self.generic_edit(
            request,
            season.timeslots,
            pk=pk,
            instance=instance,
            form_fields=(
                "start_date",
                "end_date",
                "start",
                "interval",
                "count",
            ),
            post_save_redirect=self.redirect(season.urls["edit"]),
            permission_required=True,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def delete_timeslot(self, request, season, pk, **kwargs):
        post_delete_redirect = self.redirect(season.urls["edit"] + "#timeslots-tab")
        return self.generic_delete(
            request,
            SeasonMatchTime,
            pk=pk,
            permission_required=True,
            post_delete_redirect=post_delete_redirect,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def edit_seasonexclusiondate(
        self, request, season, extra_context, pk=None, **kwargs
    ):
        instance = SeasonExclusionDate(season=season) if pk is None else None
        return self.generic_edit(
            request,
            SeasonExclusionDate,
            pk=pk,
            instance=instance,
            form_fields=("date",),
            post_save_redirect=self.redirect(season.urls["edit"]),
            permission_required=True,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def delete_seasonexclusiondate(self, request, season, pk, **kwargs):
        post_delete_redirect = self.redirect(season.urls["edit"] + "#exclusions-tab")
        return self.generic_delete(
            request,
            season.exclusions,
            pk=pk,
            permission_required=True,
            post_delete_redirect=post_delete_redirect,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def timeslots(self, request, competition, season, extra_context, **kwargs):
        return self.generic_edit_multiple(
            request,
            SeasonMatchTime,
            formset_class=SeasonMatchTimeFormSet,
            formset_kwargs={"instance": season},
            post_save_redirect_args=(competition.pk, season.pk),
            post_save_redirect="competition:season:edit",
            templates=self.template_path("timeslots.html", "season"),
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def edit_venue(self, request, season, extra_context, venue=None, **kwargs):
        if venue is None:
            venue = next_related_factory(Venue, season)
            # By default timezone should inherit from the season
            venue.timezone = season.timezone

        return self.generic_edit(
            request,
            Venue,
            instance=venue,
            related=("grounds",),
            form_class=VenueForm,
            form_kwargs={"user": request.user},
            # formset_class=GroundFormSet,
            post_save_redirect=self.redirect(season.urls["edit"]),
            permission_required=True,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def delete_venue(self, request, season, venue, extra_context=None, **kwargs):
        post_delete_redirect = self.redirect(season.urls["edit"] + "#venues-tab")
        return self.generic_delete(
            request,
            Venue,
            pk=venue.pk,
            permission_required=True,
            post_delete_redirect=post_delete_redirect,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def edit_ground(self, request, venue, extra_context, ground=None, **kwargs):
        if ground is None:
            ground = next_related_factory(Ground, venue, "venue")
            # By default timezone and geo should inherit from the venue
            ground.timezone = venue.timezone
            ground.latlng = venue.latlng

        return self.generic_edit(
            request,
            venue.grounds,
            instance=ground,
            related=(),
            form_class=GroundForm,
            form_kwargs={"user": request.user},
            post_save_redirect=self.redirect(venue.urls["edit"]),
            permission_required=True,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def delete_ground(self, request, venue, ground, extra_context=None, **kwargs):
        post_delete_redirect = self.redirect(venue.urls["edit"] + "#grounds-tab")
        return self.generic_delete(
            request,
            Ground,
            pk=ground.pk,
            permission_required=True,
            post_delete_redirect=post_delete_redirect,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def edit_division(self, request, season, extra_context, division=None, **kwargs):
        if division is None:
            division = next_related_factory(Division, season)

        return self.generic_edit(
            request,
            Division,
            instance=division,
            related=(
                "teams",
                "stages",
                "exclusions",
            ),
            form_class=DivisionForm,
            form_kwargs={"user": request.user},
            post_save_redirect=self.redirect(season.urls["edit"]),
            permission_required=True,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def delete_division(self, request, season, division, **kwargs):
        post_delete_redirect = self.redirect(season.urls["edit"] + "#divisions-tab")
        return self.generic_delete(
            request,
            Division,
            pk=division.pk,
            permission_required=True,
            post_delete_redirect=post_delete_redirect,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def edit_divisionexclusiondate(
        self, request, division, extra_context, pk=None, **kwargs
    ):
        if pk is None:
            instance = DivisionExclusionDate(division=division)
        else:
            instance = None
        return self.generic_edit(
            request,
            DivisionExclusionDate,
            pk=pk,
            instance=instance,
            form_fields=("date",),
            post_save_redirect=self.redirect(division.urls["edit"]),
            permission_required=True,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def delete_divisionexclusiondate(self, request, division, pk, **kwargs):
        post_delete_redirect = self.redirect(division.urls["edit"] + "#exclusions-tab")
        return self.generic_delete(
            request,
            division.exclusions,
            pk=pk,
            permission_required=True,
            post_delete_redirect=post_delete_redirect,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def edit_stage(self, request, division, extra_context, stage=None, **kwargs):
        if stage is None:
            stage = next_related_factory(Stage, division)

        def post_save_callback(obj):
            return self.redirect(obj.urls["edit"])

        return self.generic_edit(
            request,
            division.stages.select_related(
                "matches__home_team",
                "matches__away_team",
            ),
            pk=stage.pk,
            instance=stage,
            related=(
                "pools",
                "undecided_teams",
                "matches",
            ),
            form_class=StageForm,
            form_kwargs={"user": request.user},
            post_save_callback=post_save_callback,
            post_save_redirect=self.redirect(division.urls["edit"]),
            permission_required=True,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def delete_stage(self, request, division, stage, **kwargs):
        post_delete_redirect = self.redirect(division.urls["edit"] + "#stages-tab")
        return self.generic_delete(
            request,
            division.stages,
            pk=stage.pk,
            permission_required=True,
            post_delete_redirect=post_delete_redirect,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def edit_pool(self, request, stage, extra_context, pool=None, **kwargs):
        if pool is None:
            pool = next_related_factory(StageGroup, stage)

        def post_save_callback(obj):
            if pool:
                return self.redirect(obj.urls["edit"])

        return self.generic_edit(
            request,
            StageGroup,
            instance=pool,
            related=(),
            form_class=StageGroupForm,
            form_kwargs={"user": request.user},
            post_save_callback=post_save_callback,
            post_save_redirect=self.redirect(stage.urls["edit"]),
            permission_required=True,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def delete_pool(self, request, stage, pool, **kwargs):
        post_delete_redirect = self.redirect(stage.urls["edit"] + "#pools-tab")
        return self.generic_delete(
            request,
            stage.pools,
            pk=pool.pk,
            permission_required=True,
            post_delete_redirect=post_delete_redirect,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def edit_team(
        self,
        request,
        competition,
        season,
        division,
        extra_context,
        stage=None,
        team=None,
        **kwargs,
    ):
        """
        This view gets mounted twice, first at the division level and secondly
        at the stage level. When at the competition level it will deal with
        `Team` instances and with `UndecidedTeam` instances at the stage level.

         * raise a `Http404` when `Stage.order` == 1
        """
        related = ()

        if stage:
            model = UndecidedTeam
            post_save_redirect = self.redirect(stage.urls["edit"])
            form_class, form_kwargs = UndecidedTeamForm, {}
            if team is None:
                team = UndecidedTeam(stage=stage)
        else:
            model = Team
            post_save_redirect = self.redirect(division.urls["edit"])
            form_class, form_kwargs = TeamForm, {"division": division}
            if team is None:
                team = next_related_factory(Team, division)
            # only "real" teams have people in them
            related = ("people",)

        return self.generic_edit(
            request,
            model,
            instance=team,
            related=related,
            form_class=form_class,
            form_kwargs=form_kwargs,
            post_save_redirect=post_save_redirect,
            permission_required=True,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def perms_team(self, request, team, **kwargs):
        competition = kwargs.get("competition")
        season = kwargs.get("season")
        division = kwargs.get("division")
        return self.generic_permissions(
            request,
            Team,
            instance=team,
            post_save_redirect="admin:fixja:competition:season:" "division:team:list",
            post_save_redirect_args=(
                competition.pk,
                season.pk,
                division.pk,
            ),
            permission_required=True,
            **kwargs,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def delete_team(
        self,
        request,
        competition,
        season,
        division,
        team,
        extra_context,
        stage=None,
        **kwargs,
    ):
        if team.matches.exists():
            return HttpResponseGone("The team has matches associated with it.")

        if stage:
            manager = stage.undecided_teams
            post_delete_redirect_url = self.reverse(
                "competition:season:division:stage:edit",
                args=(competition.pk, season.pk, division.pk, stage.pk),
            )
        else:
            manager = division.teams
            post_delete_redirect_url = self.reverse(
                "competition:season:division:edit",
                args=(competition.pk, season.pk, division.pk),
            )

        return self.generic_delete(
            request,
            manager,
            pk=team.pk,
            post_delete_redirect=self.redirect(post_delete_redirect_url),
            permission_required=False,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def edit_teamassociation(self, request, team, extra_context, pk=None, **kwargs):
        if pk is None:
            instance = TeamAssociation(team=team)
        else:
            instance = get_object_or_404(team.people, pk=pk)

        return self.generic_edit(
            request,
            team.people,
            instance=instance,
            form_class=TeamAssociationForm,
            form_kwargs={
                "team": team,
            },
            post_save_redirect=self.redirect(team.urls["edit"]),
            permission_required=True,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def delete_teamassociation(self, request, team, pk, **kwargs):
        post_delete_redirect = self.redirect(team.urls["edit"] + "#people-tab")
        return self.generic_delete(
            request,
            team.people,
            pk=pk,
            permission_required=True,
            post_delete_redirect=post_delete_redirect,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def edit_seasonreferee(self, request, season, extra_context, pk=None, **kwargs):
        # FIXME this dance to establish an instance on a related object
        #       feels like it's broken to me. If provided with a related
        #       manager it should create against that? I might be over
        #       simplifying that - django.contrib.admin probably doesn't
        #       do that since they use a very flat admin (no nesting).
        if pk is None:
            instance = SeasonReferee(season=season)
        else:
            instance = get_object_or_404(season.referees, pk=pk)
        return self.generic_edit(
            request,
            season.referees,
            instance=instance,
            form_fields=("person", "club"),
            post_save_redirect=self.redirect(season.urls["edit"]),
            permission_required=True,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def delete_seasonreferee(self, request, season, pk, **kwargs):
        post_delete_redirect = self.redirect(season.urls["edit"] + "#officials-tab")
        return self.generic_delete(
            request,
            season.referees,
            pk=pk,
            permission_required=True,
            post_delete_redirect=post_delete_redirect,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def generate_draw(
        self, request, competition, season, division, stage, extra_context, **kwargs
    ):

        form_list = [
            DrawGenerationFormSet,
            DrawGenerationMatchFormSet,
        ]

        redirect_to = self.reverse(
            "competition:season:division:edit",
            args=(competition.pk, season.pk, division.pk),
        )

        view = DrawGenerationWizard.as_view(
            form_list, extra_context=extra_context, redirect_to=redirect_to, stage=stage
        )

        return view(request)

    @competition_by_pk_m
    @staff_login_required_m
    def undo_draw(self, request, division, stage, **kwargs):
        LadderSummary.objects.filter(stage=stage).delete()
        LadderEntry.objects.filter(match__stage=stage).delete()
        Match.objects.filter(stage=stage).delete()
        messages.success(request, _("Your draw has been undone."))
        return self.redirect(division.urls["edit"])

    @competition_by_pk_m
    @staff_login_required_m
    def edit_match(
        self,
        request,
        competition,
        season,
        division,
        stage,
        extra_context,
        match=None,
        **kwargs,
    ):
        if match is None:
            match = Match(stage=stage, include_in_ladder=stage.keep_ladder)

        return self.generic_edit(
            request,
            Match,
            instance=match,
            form_class=MatchEditForm,
            post_save_redirect=self.redirect(stage.urls["edit"]),
            permission_required=True,
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def delete_match(self, request, stage, match, **kwargs):
        post_delete_redirect = self.redirect(stage.urls["edit"] + "#matches-tab")
        return self.generic_delete(
            request,
            stage.matches,
            pk=match.pk,
            permission_required=True,
            post_delete_redirect=post_delete_redirect,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def edit_match_detail(self, request, stage, match, extra_context, **kwargs):
        redirect_to = reverse("admin:index")
        return super(CompetitionAdminComponent, self).edit_match_detail(
            request,
            stage=stage,
            match=match,
            extra_context=extra_context,
            redirect_to=redirect_to,
            **kwargs,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def day_runsheet(self, request, season, date, extra_context, **kwargs):
        return super(CompetitionAdminComponent, self).day_runsheet(
            request, season, date, extra_context, **kwargs
        )

    @competition_by_pk_m
    @csrf_exempt_m
    @staff_login_required_m
    def season_grid(self, request, season, mode, extra_context, **kwargs):
        # pass query string parameters along to the context
        for key, val in request.GET.items():
            extra_context.setdefault(key, val)

        return generate_fixture_grid(
            season, format=mode, request=request, extra_context=extra_context
        )

    @competition_by_pk_m
    @csrf_exempt_m
    @staff_login_required_m
    def day_grid(self, request, season, date, mode, extra_context, **kwargs):
        return generate_fixture_grid(
            season,
            format=mode,
            request=request,
            extra_context=extra_context,
            dates=[date],
        )

    @competition_by_pk_m
    @staff_login_required_m
    def progress_teams(
        self, request, competition, season, division, stage, extra_context, **kwargs
    ):
        matches = stage.matches.filter(team_needs_progressing).exclude(
            legitimate_bye_match
        )
        teams = stage.undecided_teams.all()

        if not matches:
            return HttpResponseGone()

        elif matches.filter(
            Q(home_team_undecided__isnull=False, home_team__isnull=True)
            | Q(away_team_undecided__isnull=False, away_team__isnull=True)
        ):
            follows = stage.comes_after

            ladders = collections.OrderedDict()
            if follows.pools.count():
                for pool in follows.pools.all():
                    ladders.setdefault(
                        pool, pool.ladder_summary.select_related("team__club")
                    )
            else:
                ladders.setdefault(
                    None, follows.ladder_summary.select_related("team__club")
                )

            extra_context["ladders"] = ladders

            generic_edit_kwargs = {
                "formset_class": ProgressTeamsFormSet,
                "formset_kwargs": {
                    "stage": stage,
                },
                "model_or_manager": teams.model,
                "templates": self.template_path("progress_teams.html"),
            }
        else:
            generic_edit_kwargs = {
                "formset_class": ProgressMatchesFormSet,
                "formset_kwargs": {
                    "queryset": matches,
                },
                "model_or_manager": matches.model,
                "templates": self.template_path("progress_matches.html"),
            }

        return self.generic_edit_multiple(
            request,
            post_save_redirect=self.redirect(season.urls["edit"]),
            permission_required=False,
            extra_context=extra_context,
            **generic_edit_kwargs,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def match_reschedule(self, request, competition, season, extra_context, **kwargs):
        matches = Match.objects.filter(
            match_unplayed,
            stage__division__season=season,
            stage__division__season__competition=competition,
        )

        dates = matches.dates("date", "day")
        redirect_to = season.urls["edit"]

        if request.method == "POST":
            formset = RescheduleDateFormSet(
                matches=matches, dates=dates, data=request.POST
            )
            if formset.is_valid():
                changes = formset.save()
                if changes:
                    message = (
                        ngettext(
                            "Your change to %(count)d match has been saved.",
                            "Your changes to %(count)d matches have been saved.",
                            changes,
                        )
                        % {"count": changes}
                    )
                    messages.success(request, message)
                else:
                    messages.info(request, _("No matches were rescheduled."))
                return self.redirect(redirect_to)
        else:
            formset = RescheduleDateFormSet(matches=matches, dates=dates)

        context = {
            "dates": dates,
            "formset": formset,
            "model": matches.none(),
            "cancel_url": redirect_to,
        }
        context.update(extra_context)

        templates = self.template_path("reschedule.html")
        return self.render(request, templates, context)

    # FIXME
    @competition_by_pk_m
    @staff_login_required_m
    def match_schedule(
        self,
        request,
        competition,
        season,
        date,
        extra_context,
        time=None,
        division=None,
        stage=None,
        round=None,
        **kwargs,
    ):
        where = Q(date=date, is_bye=False)

        if stage:
            manager = stage.matches
            if round:
                where &= Q(round=round)
        elif division:
            manager = division.matches
        elif season:
            manager = season.matches

        if time is not None:
            where &= Q(time=time)

        queryset = manager.filter(where).order_by(
            "stage__division__order", "round", "datetime", "play_at"
        )

        return self.generic_edit_multiple(
            request,
            queryset,
            formset_class=MatchScheduleFormSet,
            post_save_redirect=self.redirect(season.urls["edit"]),
            templates=self.template_path("match/schedule.html"),
            extra_context=extra_context,
        )

    @competition_by_pk_m
    @staff_login_required_m
    def match_results(
        self,
        request,
        competition,
        season,
        date,
        extra_context,
        division=None,
        time=None,
        **kwargs,
    ):
        redirect_to = reverse("admin:index", current_app=self.app)
        return super(CompetitionAdminComponent, self).match_results(
            request,
            competition,
            season,
            date,
            extra_context,
            division=division,
            time=time,
            redirect_to=redirect_to,
            **kwargs,
        )

    # FIXME
    @competition_by_pk_m
    @staff_login_required_m
    def match_washout(
        self, request, competition, season, date, extra_context, time=None, **kwargs
    ):
        matches = Match.objects.filter(
            stage__division__season__id=season.pk,
            stage__division__season__competition__id=competition.pk,
            date=date,
        ).order_by("stage__division__order", "is_bye", "datetime", "play_at")

        # ensure only matches with progressed teams are able to be updated
        matches = matches.filter(match_unplayed)

        # FIXME too complex, be verbose so we can all read and understand it
        if time is not None:
            base = Q(time=time)
            bye_kwargs = matches.filter(time=time).values("stage", "round")
            time_or_byes = functools.reduce(
                operator.or_, [Q(time__isnull=True, **kw) for kw in bye_kwargs], base
            )
            matches = matches.filter(time_or_byes)

        return self.generic_edit_multiple(
            request,
            matches,
            formset_class=MatchWashoutFormSet,
            templates=self.template_path("match_washout.html"),
            post_save_redirect=self.redirect(season.urls["edit"]),
            extra_context=extra_context,
        )

    @staff_login_required_m
    def scorecard_report(self, request, **extra_context):
        from .wizards import FilterForm, SeasonForm, scorecardwizard_factory

        ScorecardWizard = scorecardwizard_factory(app=self, extra_context=extra_context)
        wizard = ScorecardWizard.as_view(form_list=[SeasonForm, FilterForm])
        return wizard(request)

    @competition_by_pk_m
    @staff_login_required_m
    def scorecards(
        self,
        request,
        competition,
        season,
        mode,
        stage=None,
        date=None,
        time=None,
        **kwargs,
    ):
        if stage is not None:
            if stage.matches_needing_printing.count():
                matches = stage.matches_needing_printing.all()
            else:
                matches = stage.matches.all()
        else:
            matches = season.matches.all()

        # legacy/undocumented, remove this?
        if "d" in request.GET:
            matches = matches.filter(stage__division__slug=request.GET["d"])

        if date is not None:
            matches = matches.filter(date=date)

        if time is not None:
            matches = matches.filter(time=time)

        matches = matches.exclude(legitimate_bye_match)

        extra_context = {
            "competition": competition,
            "season": season,
            "date": date,
            "time": time,
            "stage": stage,
            "filtered": round is not None,
            "round": round,
        }

        templates = self.template_path("scorecards.html", competition.slug, season.slug)

        if mode == "pdf":
            kw = {
                "match_pks": [pk for pk in matches.values_list("pk", flat=True)],
                "templates": templates,
                "extra_context": extra_context,
            }
            if stage is not None:
                kw["stage_pk"] = stage.pk
            if hasattr(request, "tenant"):
                kw["_schema_name"] = request.tenant.schema_name
                kw["base_url"] = request.build_absolute_uri("/")

            result = generate_pdf_scorecards.delay(**kw)

            reverse_kwargs = {
                "competition_id": competition.pk,
                "season_id": season.pk,
                "result_id": result.id,
            }

            redirect_to = self.reverse(
                "competition:season:scorecards-async", kwargs=reverse_kwargs
            )
            response = self.redirect(redirect_to)
        else:
            output = generate_scorecards(matches, templates, mode, extra_context, stage)
            response = HttpResponse(output)

        return response

    @competition_by_pk_m
    @staff_login_required_m
    def scorecards_async(self, request, result_id, extra_context, **kwargs):
        result = generate_pdf_scorecards.AsyncResult(result_id)

        if result.ready():
            data = result.wait()
            return HttpResponse(base64.b64decode(data), content_type="application/pdf")

        templates = self.template_path("wait.html", "scorecards")
        response = self.render(request, templates, extra_context)
        response["Refresh"] = SCORECARD_PDF_WAIT
        return response

    @competition_by_pk_m
    @staff_login_required_m
    def highest_point_scorer(self, request, division, extra_context, **kwargs):
        def _get_clause(field, aggregate=Sum):
            """
            Local function to produce a Aggregate(Case(When())) instance which
            can be used to extract individual totals for the division.
            """
            return aggregate(
                Case(When(statistics__match__stage__division=division, then=F(field)))
            )

        people = (
            Person.objects.select_related("club")
            .annotate(
                played=_get_clause("statistics__played"),
                points=_get_clause("statistics__points"),
                mvp=_get_clause("statistics__mvp"),
            )
            .exclude(played=None)
        )

        scorers = people.order_by("-points", "played")
        mvp = people.order_by("-mvp", "played")

        context = {
            "scorers": scorers,
            "mvp": mvp,
        }
        context.update(extra_context)

        templates = self.template_path("scorers.html")
        return self.render(request, templates, context)

    # from RegistrationBase

    @staff_login_required_m
    def list_clubs(self, request, **extra_context):
        return self.generic_list(
            request,
            Club,
            paginate_by=0,
            permission_required=True,
            extra_context=extra_context,
        )

    @registration
    def edit_club(self, request, extra_context, club=None, **kwargs):
        templates = {
            Team: "tournamentcontrol/competition/admin/club/team.inc.html",
        }
        extra_context.setdefault("templates", templates)
        return self.generic_edit(
            request,
            Club,
            instance=club,
            form_fields=(
                "title",
                "short_title",
                "abbreviation",
                "email",
                "website",
                "twitter",
                "facebook",
                "youtube",
                "primary",
                "primary_position",
                "slug",
                "slug_locked",
            ),
            related=(
                "members",
                "teams",
            ),
            permission_required=True,
            extra_context=extra_context,
        )

    @staff_login_required_m
    def delete_club(self, request, club_id, **kwargs):
        return self.generic_delete(request, Club, pk=club_id, permission_required=True)

    @registration
    @staff_login_required_m
    def edit_person(self, request, club, extra_context, person=None, **kwargs):
        if person is None:
            person = Person(club=club)
        # todo
        readonly_fields = (
            "first_name",
            "last_name",
            "gender",
            "date_of_birth",
        )
        return self.generic_edit(
            request,
            club.members,
            instance=person,
            form_class=PersonEditForm,
            related=("statistics",),
            post_save_redirect=self.redirect(person.club.urls["edit"]),
            permission_required=True,
            extra_context=extra_context,
        )

    @registration
    @staff_login_required_m
    def merge_person(self, request, club, extra_context, person, **kwargs):
        return self.generic_edit(
            request,
            club.members,
            instance=person,
            form_class=PersonMergeForm,
            related=(),
            post_save_redirect=self.redirect(person.club.urls["edit"]),
            permission_required=True,
            extra_context=extra_context,
        )

    @registration
    @staff_login_required_m
    def delete_person(self, request, club, person, extra_context, **kwargs):
        now = timezone.now().replace(second=0, microsecond=0)
        queryset = Person.objects.filter(club=club)

        # Ensure the person does not have any statistical data recorded and
        # that we're not attempting to delete someone mid-season.
        queryset = queryset.exclude(statistics__isnull=False)
        queryset = queryset.exclude(
            teamassociation__team__division__season__start_date__lt=now
        )

        return self.generic_delete(
            request,
            queryset,
            pk=person.pk,
            post_delete_redirect=self.redirect(club.urls["edit"]),
            permission_required=True,
            extra_context=extra_context,
        )

    @registration
    def edit_clubassociation(
        self, request, club, extra_context, clubassociation_id=None, **kwargs
    ):
        return self.generic_edit(
            request,
            ClubAssociation,
            pk=clubassociation_id,
            form_class=ClubAssociationForm,
            form_kwargs={
                "user": request.user,
                "club": club,
            },
            permission_required=True,
            extra_context=extra_context,
        )

    @registration
    def officials(self, request, club, season, extra_context, **kwargs):
        # Read only is presently applied to the entire application, but we
        # really need to be able to provide granular control for specific
        # seasons.  FIXME
        read_only, scheduled_read_only = self.read_only
        if season.complete:
            if not request.user.is_superuser:
                read_only = True

        if read_only:
            messages.add_message(
                request,
                messages.INFO,
                _("This season is now read-only to club administrators."),
            )
        elif scheduled_read_only:
            sro = {
                "time": scheduled_read_only.strftime("%H:%M"),
                "date": scheduled_read_only.strftime("%d/%m/%Y"),
            }
            messages.add_message(
                request,
                messages.WARNING,
                _(
                    "This season will become read-only to club administrators "
                    "at %(time)s on %(date)s."
                )
                % sro,
            )

        if read_only:
            templates = self.template_path("officials_list.html")
            return self.render(request, templates, extra_context)

        templates = self.template_path("officials.html")
        return self.generic_edit_multiple(
            request,
            SeasonAssociation,
            queryset=season.officials.filter(club=club),
            formset_class=SeasonAssociationFormSet,
            formset_kwargs={
                "user": request.user,
                "club": club,
                "season": season,
            },
            post_save_redirect="club-teams",
            post_save_redirect_kwargs={"club_id": club.pk},
            templates=templates,
            extra_context=extra_context,
        )

    @registration
    def edit_team_members(self, request, club, team, extra_context, **kwargs):
        statistics = SimpleScoreMatchStatistic.objects.filter(
            Q(player__teamassociation__team=team) & Q(match__home_team=team)
            | Q(match__away_team=team)
        )

        # FIXME this is a hangover from before we had permissions on generics
        read_only, scheduled_read_only = request.user.is_superuser, False

        # Read only is presently applied to the entire application, but we
        # really need to be able to provide granular control for specific
        # seasons.  FIXME
        if team.division.season.complete:
            if not request.user.is_superuser:
                read_only = True

        if read_only:
            messages.add_message(
                request,
                messages.INFO,
                _("This team is now read-only to club administrators."),
            )
        elif scheduled_read_only:
            sro = {
                "time": scheduled_read_only.strftime("%H:%M"),
                "date": scheduled_read_only.strftime("%d/%m/%Y"),
            }
            messages.add_message(
                request,
                messages.WARNING,
                _(
                    "This team will become read-only to club administrators at "
                    "%(time)s on %(date)s."
                )
                % sro,
            )

        if read_only and not request.user.is_superuser:
            templates = self.template_path("team_members.html")
            return self.render(request, templates, extra_context)

        return self.generic_edit_multiple(
            request,
            TeamAssociation,
            formset_class=TeamAssociationFormSet,
            formset_kwargs={"user": request.user, "instance": team},
            post_save_redirect="club-teams",
            post_save_redirect_kwargs={"club_id": club.pk},
            permission_required=True,
            extra_context=extra_context,
        )

    @registration
    @staff_login_required_m
    def registration_form(
        self, request, club, season, teams, competition, mode, extra_context, **kwargs
    ):
        data = collections.OrderedDict()

        for team in teams:
            for pa in team.people.filter(is_player=True, person__isnull=False):
                data.setdefault(team, collections.OrderedDict()).setdefault(
                    "players", []
                ).append(
                    (
                        pa.number or "",
                        pa.person.first_name.encode("utf-8").strip(),
                        pa.person.last_name.encode("utf-8").strip(),
                        pa.person.date_of_birth,
                        pa.person.email,
                    )
                )

            for pa in team.people.filter(
                roles__isnull=False, person__isnull=False
            ).order_by("roles__id", "person__last_name", "person__first_name"):
                for role in pa.roles.all():
                    data.setdefault(team, collections.OrderedDict()).setdefault(
                        "staff", collections.OrderedDict()
                    ).setdefault(role.name, set()).add(
                        (
                            pa.person.first_name.encode("utf-8").strip(),
                            pa.person.last_name.encode("utf-8").strip(),
                            pa.person.date_of_birth,
                            pa.person.email,
                        )
                    )

        extra_context["data"] = data

        templates = [
            "tournamentcontrol/rego/%s/%s/registration_form.html"
            % (competition.slug, season.slug),
            "tournamentcontrol/rego/%s/registration_form.html" % (competition.slug,),
            "tournamentcontrol/rego/registration_form.html",
        ]

        res = TemplateResponse(request, templates, extra_context)

        if mode == "pdf":
            pdf = prince(res.render().content)
            return HttpResponse(pdf, content_type="application/pdf")

        return res



from __future__ import unicode_literals

from datetime import datetime
from os.path import dirname, join

from django.test.utils import modify_settings, override_settings
from django.utils import timezone
from django.utils.encoding import smart_str
from test_plus import TestCase
from touchtechnology.common.tests.test_timezone_forms import TestForm1


class TimeZoneTests(TestCase):

    maxDiff = None

    def get_html(self, filename):
        with open(join(dirname(__file__), "test_html", filename)) as fp:
            html = fp.read()
        return smart_str(html)

    @override_settings(USE_TZ=False)
    def test_select_date_time_field_empty(self):
        data = {
            "timestamp_0": "4",
            "timestamp_1": "7",
            "timestamp_2": "2013",
            "timestamp_3": "",
            "timestamp_4": "",
        }

        # with only date fields completed validation should fail
        form = TestForm1(data=data)
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors["timestamp"], ["Please enter a valid date and time."]
        )

        data.update(
            {"timestamp_3": "7", "timestamp_4": "49",}
        )

        # time fields added should allow validation to succeed
        form = TestForm1(data=data)
        self.assertTrue(form.is_valid())

        # verify the cleaned value from user input is a datetime
        # object as we would expect it.
        timestamp = form.cleaned_data.get("timestamp")
        self.assertEqual(datetime(2013, 7, 4, 7, 49), timestamp)

    @override_settings(USE_TZ=True, TIME_ZONE="Australia/Sydney")
    def test_select_date_time_field_empty_tz(self):
        data = {
            "timestamp_0": "4",
            "timestamp_1": "7",
            "timestamp_2": "2013",
            "timestamp_3": "",
            "timestamp_4": "",
            "timestamp_5": "Europe/London",
        }

        # with only date fields completed validation should fail
        form = TestForm1(data=data)
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors["timestamp"], ["Please enter a valid date and time."]
        )

        data.update(
            {"timestamp_3": "7", "timestamp_4": "49",}
        )

        # time fields added should allow validation to succeed
        form = TestForm1(data=data)
        self.assertTrue(form.is_valid())

        # verify the cleaned value from user input is a datetime
        # object as we would expect it.
        timestamp = form.cleaned_data.get("timestamp")
        self.assertEqual(
            datetime(2013, 7, 4, 16, 49),
            timezone.make_naive(timestamp, timezone.get_current_timezone()),
        )

    @override_settings(USE_TZ=False)
    def test_select_date_time_field_initial(self):
        timestamp = datetime(2013, 3, 24, 14, 30)

        initial = {
            "timestamp": timestamp,
        }
        form = TestForm1(initial=initial)

        # check that the HTML output is as expected for a naked
        # form renderer.
        html1 = smart_str(form)
        html2 = self.get_html("select_date_time_field_initial.html")
        self.assertHTMLEqual(html1, html2)

    @override_settings(USE_TZ=True, TIME_ZONE="Australia/Sydney")
    def test_select_date_time_field_initial_tz(self):
        timestamp = timezone.make_aware(
            datetime(2013, 3, 24, 14, 30), timezone.get_current_timezone()
        )

        initial = {
            "timestamp": timestamp,
        }
        form = TestForm1(initial=initial)

        # check that the HTML output is as expected for a naked
        # form renderer.
        html1 = smart_str(form)
        html2 = self.get_html("select_date_time_field_initial_tz.html")
        self.assertHTMLEqual(html1, html2)


@override_settings(ROOT_URLCONF="example_app.urls")
@modify_settings(
    MIDDLEWARE={"append": "touchtechnology.common.middleware.timezone_middleware",}
)
class TestTimezoneMiddleware(TestCase):
    def test_set_timezone(self):
        "Ensure time zone is correct before, during and after a set request"
        self.assertEqual(
            timezone.get_current_timezone_name(), timezone.get_default_timezone_name()
        )

        data = {
            "timezone": "Australia/Sydney",
        }
        self.post("set-timezone", data=data, follow=True)
        self.assertResponseContains(
            '<option value="Australia/Sydney" selected>Sydney</option>'
        )

        self.assertEqual(
            timezone.get_current_timezone_name(), timezone.get_default_timezone_name()
        )



from django.test import TestCase

from tournamentcontrol.competition.models import LadderEntry, LadderSummary
from tournamentcontrol.competition.tests import factories


class SignalHandlerTests(TestCase):
    def test_ladder_entry_match_no_score(self):
        factories.MatchFactory.create()
        self.assertQuerysetEqual(LadderEntry.objects.all(), LadderEntry.objects.none())

    def test_ladder_entry_match_with_score(self):
        factories.MatchFactory.create(home_team_score=5, away_team_score=2)
        self.assertCountEqual(
            [
                (1, 0, 0, 5, 2, 3, 3),
                (0, 1, 0, 2, 5, -3, 3),
            ],
            LadderEntry.objects.values_list(
                "win", "loss", "draw", "score_for", "score_against", "diff", "margin"
            ),
        )

    def test_ladder_summary_3_match_series(self):
        # seed a match which will produce all required relations
        match = factories.MatchFactory.create(home_team_score=5, away_team_score=2)

        # build two more matches with known results to produce LadderSummary
        factories.MatchFactory.create(
            home_team=match.home_team,
            home_team_score=3,
            away_team=match.away_team,
            away_team_score=4,
            stage=match.stage,
        )
        factories.MatchFactory.create(
            home_team=match.home_team,
            home_team_score=2,
            away_team=match.away_team,
            away_team_score=2,
            stage=match.stage,
        )

        self.assertCountEqual(
            [
                (3, 1, 1, 1, 10, 8, 2),
                (3, 1, 1, 1, 8, 10, -2),
            ],
            LadderSummary.objects.values_list(
                "played",
                "win",
                "loss",
                "draw",
                "score_for",
                "score_against",
                "difference",
            ),
        )



from django.contrib.sitemaps import Sitemap
from django.db.models import Q
from touchtechnology.common.models import SitemapNode
from touchtechnology.common.utils import create_exclude_filter


class NodeSitemap(Sitemap):

    def items(self):
        nodes = SitemapNode._tree_manager.select_related('content_type')

        nodes_hidden_from_sitemap = nodes.filter(
            Q(hidden_from_sitemap=True) | Q(enabled=False))
        hidden_from_sitemap = create_exclude_filter(nodes_hidden_from_sitemap)
        nodes = nodes.exclude(hidden_from_sitemap)

        nodes = nodes.order_by('tree_id', 'lft')

        # flatten the list of nodes to a list
        tree = list(nodes)

        # insert faux SitemapNode objects into the tree where possible
        placeholders = nodes.filter(content_type__app_label='content',
                                    content_type__model='placeholder')
        for app_node in sorted(placeholders, reverse=True):
            site = app_node.object.site(app_node)
            if hasattr(site, 'sitemapnodes'):
                index = tree.index(app_node) + 1
                tree = tree[:index] + list(site.sitemapnodes()) + tree[index:]

        return tree

    def priority(self, obj):
        if hasattr(obj, 'priority'):
            return getattr(obj, 'priority')
        return 0.5

    def changefreq(self, obj):
        if hasattr(obj, 'changefreq'):
            return getattr(obj, 'changefreq')
        return 'weekly'



# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2017-09-25 08:13
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("common", "0003_drop_pythonpackage"),
    ]

    operations = [
        migrations.AddField(
            model_name="sitemapnode",
            name="kwargs",
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
    ]



import logging

from django.template import Library
from django.utils.safestring import mark_safe
from touchtechnology.content.models import Chunk

logger = logging.getLogger(__name__)
register = Library()


@register.simple_tag
def chunk(slug):
    instance, created = Chunk.objects.get_or_create(slug=slug)
    if created:
        logger.debug("Created %r" % instance)
    return mark_safe(instance.copy)



from django import forms
from django.utils.translation import gettext_lazy as _


def boolean_coerce(value):
    if value in {1, "1"}:
        return True
    if value in {0, "0"}:
        return False


class BooleanChoiceField(forms.TypedChoiceField):
    widget = forms.Select

    def __init__(self, *args, **kwargs):
        defaults = {
            "choices": [
                ("1", _("Yes")),
                ("0", _("No")),
            ],
            "coerce": boolean_coerce,
            "required": True,
        }
        defaults.update(kwargs)
        super(BooleanChoiceField, self).__init__(*args, **defaults)

    def prepare_value(self, value):
        if value is not None:
            return str(int(value))
